/* ================== Doxygen Comment ======================================= */
/**
       \file
       \brief This module contains some usual functions definitions to handle dates.


  TEAM_UNIQUE HEADER:

       Equipe Interaction Rayonnement Nuages - Laboratoire d'Optique Atmospherique
       Universite des Sciences et Technologies de Lille, Villeneuve d'Ascq

  PROGRAMMERS:

       Jerome Riedi  (jerome dot riedi at univ-lille1 dot fr)
       Francois Thieuleux (francois dot thieuleux at univ-lille1 dot fr)

  CORRESPONDING AUTHOR

       Jerome Riedi  (jerome dot riedi at univ-lille1 dot fr)

  REFERENCES AND CREDITS

       This work has been supported by USTL, CNRS and CNES.
       All credits go to Laboratoire d'Optique Atmospherique (USTL)

  COPYRIGHT (c) 2004-2012 by 

       Laboratoire d'Optique Atmospherique (UMR8518 CNRS/USTL)
       Batiment P5, USTL
       59 655 Villeneuve d'Ascq Cedex
       FRANCE
       http://www-loa.univ-lille1.fr

  LICENSE

       This program is free software; you can redistribute it and/or
       modify it under the terms of the GNU General Public License
       as published by the Free Software Foundation; either version 2
       of the License, or any later version.

       This program is distributed in the hope that it will be useful,
       but WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
       GNU General Public License for more details.

       You should have received a copy of the GNU General Public License
       along with this program; if not, write to the Free Software Foundation,
       Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
/* ================== Doxygen End Comment =================================== */
#ifdef __cplusplus
#error This source file is not C++ but rather C. Please use a C-compiler
#endif

#include "com_fct_date.h"

/* ==== macros ============================================================== */
/* ==== constants =========================================================== */
/* ==== types =============================================================== */
/* ==== structures ========================================================== */
/* ==== private variables =================================================== */
/* ==== private functions =================================================== */
/* ==== internal public functions =========================================== */


/*==============================================================================*/
/*             UTILS METEO                                                      */
/*==============================================================================*/


/*--------------------------------- -----------------------------------*/
/**
 *  \brief 
 *
 *  julday function returns the Julian Day number that begins at noon
 *  of the calendar date specified by month mm, day id, and year iyyy, 
 *  all integer variables. Positive year signifies A.D.; negative B.C.
 *  Remember that the year after 1 B.C. was 1 A.D. .
 *
 *  Astronomers number each 24-hour period , starting and ending at noon,
 *  with a unique integer, the Julian Day number. Julian Day Zero was a
 *  very long time ago; a convenient reference point is that Julian Day 
 *  2440000 began at noon of May 23, 1968. If you know the Julian Day
 *  Number that begins at noon of a given calendar date, then the day of
 *  the week of that date is obtained by adding 1 and taking the result
 *  modulo base 7; a zero answer corresponds to Sunday, 1 to Monday, ...,
 *  6 to Saturday.
 *
 */
/*--------------------------------- -----------------------------------*/
int julday(int const mm, int const id, int const iyyy)
{
  int const IGREG = 15+31*(10+12*1582);
  int ja;
  int jm;
  int jy;
  int jday;

  jy=iyyy;
  if (jy<0) { jy=jy+1; }
  if (mm>2)
  {
    jm=mm+1;
  }
  else
  {
    jy=jy-1;
    jm=mm+13;
  }

  jday=(int)(365.25*jy)+(int)(30.6001*jm)+id+1720995;

  if (id+31*(mm+12*iyyy)>=IGREG)
  {
    ja=(int)(0.01*jy);
    jday=jday+2-ja+(int)(0.25*ja);
  }
  return jday;

} /* julday */


/*--------------------------------- -----------------------------------*/
/**
 *  \brief caldat function computes month, day and year corresponding to
 *         a given Julian Day Number (see function julian defined above).
 *
 * caldat function is the inverse of the function julday defined above.
 * Here julian is input as a Julian Day Number, and the routine outputs
 * mm, id, and iyyy as the month, day, and year on which the specified 
 * Julian Day started at noon.
 *
 *  Astronomers number each 24-hour period , starting and ending at noon,
 *  with a unique integer, the Julian Day number. Julian Day Zero was a
 *  very long time ago; a convenient reference point is that Julian Day 
 *  2440000 began at noon of May 23, 1968. If you know the Julian Day
 *  Number that begins at noon of a given calendar date, then the day of
 *  the week of that date is obtained by adding 1 and taking the result
 *  modulo base 7; a zero answer corresponds to Sunday, 1 to Monday, ...,
 *  6 to Saturday.
 *
 */
/*--------------------------------- -----------------------------------*/
void caldat(int const julian, int *mm, int *id, int *iyyy)
{
  int const IGREG = 2299161;
  int ja;
  int jalpha;
  int jb;
  int jc;
  int jd;
  int je;

  if(julian>=IGREG)
  {
    jalpha=(int)(((julian-1867216)-0.25)/36524.25);
    ja=julian+1+jalpha-(int)(0.25*jalpha);
  }
  else
  {
    ja=julian;
  }
  jb=ja+1524;
  jc=(int)(6680.+((jb-2439870)-122.1)/365.25);
  jd=365*jc+(int)(0.25*jc);
  je=(int)((jb-jd)/30.6001);
  (*id)=jb-jd-(int)(30.6001*je);
  (*mm)=je-1;

  if((*mm)>12) { (*mm)=(*mm)-12; }
  (*iyyy)=jc-4715;
  if((*mm)>2) { (*iyyy)=(*iyyy)-1; }
  if((*iyyy)<=0) { (*iyyy)=(*iyyy)-1; }

} /* caldat */


int cds2jmahms(int cdsday,int cdsmsec,int *jour,int *mois,int *annee,int *heure,int *min,int *sec)
{
  int tmp,a,m,j;

  caldat(julday(1993,1,1)+cdsday,&j,&m,&a);
  *annee=a;
  *mois=m;
  *jour=j;

  *heure = cdsmsec/3600000;
  tmp    = cdsmsec - *heure*3600000;
  *min   = tmp/60000;
  tmp    = tmp - *min*60000;
  *sec   = tmp/1000;

  return(0);
} /* Fin de la fonction cds2jmahms */


/* ==== entry points ======================================================== */
/* ==== public variables ==================================================== */
