/* ================== Doxygen Comment ======================================= */
/**
       \file
       \brief This module contains some basic error reporting variadic macros.


  TEAM_UNIQUE HEADER:

       Equipe Interaction Rayonnement Nuages - Laboratoire d'Optique Atmospherique
       Universite des Sciences et Technologies de Lille, Villeneuve d'Ascq

  PROGRAMMERS:

       Jerome Riedi  (jerome dot riedi at univ-lille1 dot fr)
       Francois Thieuleux (francois dot thieuleux at univ-lille1 dot fr)

  CORRESPONDING AUTHOR

       Jerome Riedi  (jerome dot riedi at univ-lille1 dot fr)

  REFERENCES AND CREDITS

       This work has been supported by USTL, CNRS and CNES.
       All credits go to Laboratoire d'Optique Atmospherique (USTL)

  COPYRIGHT (c) 2004-2012 by 

       Laboratoire d'Optique Atmospherique (UMR8518 CNRS/USTL)
       Batiment P5, USTL
       59 655 Villeneuve d'Ascq Cedex
       FRANCE
       http://www-loa.univ-lille1.fr

  LICENSE

       This program is free software; you can redistribute it and/or
       modify it under the terms of the GNU General Public License
       as published by the Free Software Foundation; either version 2
       of the License, or any later version.

       This program is distributed in the hope that it will be useful,
       but WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
       GNU General Public License for more details.

       You should have received a copy of the GNU General Public License
       along with this program; if not, write to the Free Software Foundation,
       Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
/* ================== Doxygen End Comment =================================== */
#ifndef H_COM_FCT_ERRORS_H
#define H_COM_FCT_ERRORS_H

#include <stdio.h>  /* standard symbolic constants and types */
#include <stdlib.h> /* Standard library definitions */

#include <stdarg.h> /* Handle variable argument list */

#include "verbose.h"

/* Application name definition */
#include "appname.h"

/* ==== constants =========================================================== */
/* ==== types =============================================================== */
/* ==== structures ========================================================== */
/* ==== internal public functions =========================================== */
/* ==== public variables ==================================================== */

//extern int have_error;

/* ==== macros ============================================================== */

/* Fatal error - terminate execution immediately.  Does not return.  */
/**
 *
 * \brief fatal macro print a fatal message, and terminate execution immediately. It does not return.
 *
 * fatal macro print on standard error a fatal message. It then terminates execution
 * immediately, by calling exit() standard library function.
 * This is the worst level of dialout.
 * This is done whatever the level of verbosity requested by user (see common/verbose.h).
 *
 */
#define fatal(...) \
do {\
  fprintf(stderr, "%s : Fatal   : ", program_name);\
  fprintf(stderr, __VA_ARGS__);\
  fprintf(stderr, "        %s:%d(%s)\n", __FILE__, __LINE__, __FUNCTION__);\
  exit (EXIT_FAILURE);\
} while(0)
                           

/**
 *
 * \brief debug macro print on standard error a message for debugging purposes.
 *
 * warning macro print on standard error a message for debugging purposes,
 * but only if the verbose level is equal to DIAGNOSTIC_VERBOSE (see common/verbose.h).
 *
 */
#define debug(...) \
do {\
  if (verbose == DIAGNOSTIC_VERBOSE)\
  {\
    fprintf(stderr, "%s : Debug   : ", program_name);\
    fprintf(stderr, __VA_ARGS__);\
    fprintf(stderr, "        %s:%d(%s)\n", __FILE__, __LINE__, __FUNCTION__);\
  }\
} while(0)


/**
 *
 * \brief debugn macro print on standard error a message for debugging purposes,
 *        and print a supplementary '\n' character.
 *
 * warning macro print on standard error a message for debugging purposes,
 * but only if the verbose level is equal to DIAGNOSTIC_VERBOSE (see common/verbose.h).
 * This macro is identical to previous debug macro, but print a supplementary '\n' character.
 *
 */
#define debugn(...) \
do \
{\
  if (verbose == DIAGNOSTIC_VERBOSE)\
  {\
    fprintf(stderr, "%s : Debug   : ", program_name);\
    fprintf(stderr, __VA_ARGS__);\
    fprintf(stderr, "        %s:%d(%s)", __FILE__, __LINE__, __FUNCTION__);\
    fprintf(stderr, "\n");\
  }\
} while(0)

                           
/**
 *
 * \brief error macro print on standard error a message error,
 *        whatever the level verbosity (see common/verbose.h).
 *
 * error macro print on standard error a message error,
 * whatever the level of verbosity (see common/verbose.h).
 *
 */
#define error(...) \
do \
{\
  fprintf(stderr, "%s : Error   : ", program_name);\
  fprintf(stderr, __VA_ARGS__);\
  fprintf(stderr, "        %s:%d(%s)\n", __FILE__, __LINE__, __FUNCTION__);\
} while(0)


/**
 *
 * \brief errorn macro print on standard error a message error, and print a
 *        supplementary '\n' character, whatever the level verbosity (see common/verbose.h).
 *
 * errorn macro print on standard error a message error,
 * whatever the level of verbosity (see common/verbose.h).
 * This macro is identical to previous error macro, but print a supplementary '\n' character.
 *
 */
#define errorn(...) \
do \
{\
  fprintf(stderr, "%s : Error   : ", program_name);\
  fprintf(stderr, __VA_ARGS__);\
  fprintf(stderr, "        %s:%d(%s)", __FILE__, __LINE__, __FUNCTION__);\
  fprintf(stderr, "\n");\
} while(0)


/**
 *
 * \brief warning macro print on standard error a warning message,
 *        only if verbose level is greater or equal to NORMAL_VERBOSE,
 *
 * warning macro print on standard error a warning message, if verbose level is
 * greater or equal to NORMAL_VERBOSE (see common/verbose.h). Output is producted, but there may have problems.
 *
 */
#define warning(...) \
do \
{\
  if (verbose >= NORMAL_VERBOSE)\
  {\
    fprintf(stderr, "%s : Warning : ", program_name);\
    fprintf(stderr, __VA_ARGS__);\
    fprintf(stderr, "\n");\
  }\
} while(0)

                           
/**
 *
 * \brief warningn macro print on standard error a warning message,
 *        only if verbose level is greater or equal to NORMAL_VERBOSE,
 *        and print a supplementary '\n' character.
 *
 * warningn macro print on standard error a warning message, if verbose level is
 * greater or equal to NORMAL_VERBOSE (see common/verbose.h). Output is producted, but there may have problems.
 * This macro is identical to previous warning macro, but print a supplementary '\n' character.
 *
 */
#define warningn(...) \
do \
{\
  if (verbose >= NORMAL_VERBOSE)\
  {\
    fprintf(stderr, "%s : Warning : ", program_name);\
    fprintf(stderr, __VA_ARGS__);\
    fprintf(stderr, "\n\n");\
  }\
} while(0)


/**
 *
 * \brief info macro print on standard output an informational message,
 *        only if verbose level is greater or equal to NORMAL_VERBOSE,
 *
 * info macro print on standard output an informational message, if verbose level is
 * greater or equal to NORMAL_VERBOSE (see common/verbose.h). Output is producted.
 *
 */
#define info(...) \
do \
{\
  if (verbose >= NORMAL_VERBOSE)\
  {\
    fprintf(stderr, "%s : Info    : ", program_name);\
    fprintf(stderr, __VA_ARGS__);\
    fprintf(stderr, "\n");\
  }\
} while(0)

/**
 *
 * \brief warningn macro print on standard output an informational message,
 *        only if verbose level is greater or equal to NORMAL_VERBOSE,
 *        and print a supplementary '\n' character.
 *
 * infon macro print on standard output an informational message, if verbose level is
 * greater or equal to NORMAL_VERBOSE (see common/verbose.h). Output is producted.
 * This macro is identical to previous info macro, but print a supplementary '\n' character.
 *
 */
#define infon(...) \
do \
{\
  if (verbose >= NORMAL_VERBOSE)\
  {\
    fprintf(stderr, "%s : Info    : ", program_name);\
    fprintf(stderr, __VA_ARGS__);\
    fprintf(stderr, "\n\n");\
  }\
} while(0)

/* ==== entry points ======================================================== */
/* ==== inline public functions ============================================= */

#endif /* H_COM_FCT_ERRORS_H */
