/* ================== Doxygen Comment ======================================= */
/**
       \file
       \brief This module contains string management functions definitions.


  TEAM_UNIQUE HEADER:

       Equipe Interaction Rayonnement Nuages - Laboratoire d'Optique Atmospherique
       Universite des Sciences et Technologies de Lille, Villeneuve d'Ascq

  PROGRAMMERS:

       Jerome Riedi  (jerome dot riedi at univ-lille1 dot fr)
       Francois Thieuleux (francois dot thieuleux at univ-lille1 dot fr)

  CORRESPONDING AUTHOR

       Jerome Riedi  (jerome dot riedi at univ-lille1 dot fr)

  REFERENCES AND CREDITS

       This work has been supported by USTL, CNRS and CNES.
       All credits go to Laboratoire d'Optique Atmospherique (USTL)

  COPYRIGHT (c) 2004-2012 by 

       Laboratoire d'Optique Atmospherique (UMR8518 CNRS/USTL)
       Batiment P5, USTL
       59 655 Villeneuve d'Ascq Cedex
       FRANCE
       http://www-loa.univ-lille1.fr

  LICENSE

       This program is free software; you can redistribute it and/or
       modify it under the terms of the GNU General Public License
       as published by the Free Software Foundation; either version 2
       of the License, or any later version.

       This program is distributed in the hope that it will be useful,
       but WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
       GNU General Public License for more details.

       You should have received a copy of the GNU General Public License
       along with this program; if not, write to the Free Software Foundation,
       Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
/* ================== Doxygen End Comment =================================== */
#ifdef __cplusplus
#error This source file is not C++ but rather C. Please use a C-compiler
#endif

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common/verbose.h"
#include "common/com_fct_errors.h"
#include "common/com_fct_xalloc.h"
#include "common/com_fct_string.h"

/* ==== macros ============================================================== */
/* ==== constants =========================================================== */
/* ==== types =============================================================== */
/* ==== structures ========================================================== */
/* ==== private variables =================================================== */
/* ==== private functions =================================================== */
/* ==== internal public functions =========================================== */
/* ==== entry points ======================================================== */
/* ==== public variables ==================================================== */

/* ================== Doxygen Comment ======================================= */
/* Return a filename from complete path string */
/* ================== Doxygen End Comment =================================== */
char *get_name_from_path(const char *pathname)
{
  char *fname = strrchr(pathname, '/');
  return fname == NULL ? (char*)pathname : ++fname;

} /* get_name_from_path */



/* ================== Doxygen Comment ======================================= */
/**
 *  \brief str_split function parse character string by use of a character separator.
 *
 *  str_split function parse character string given by s by mean of a
 *  character separator given by ct, and return every sub-string in an array
 *  terminated by NULL.
 *
 *  While strtok return non null address (1), we grow up size array by 
 *  one character (2), in which we keep address returned by strtok (3).
 *
 *  From http://nicolasj.developpez.com/articles/libc/string
 *  MODIF JMN : add nb_tokens in output parameters             
 *
 *
 *
 */
/* ================== Doxygen End Comment =================================== */
char **str_split(char *s, const char *ct, size_t *nb_tokens)
{
  char **tab = NULL;
  int i = -1;

  if ((s != NULL) && (ct != NULL))
  {
    char *cs = NULL;
    size_t size = 1;

/* (1) */
    for (i = 0; (cs = strtok (s, ct)); i++)
    {
      if ((int)size <= i + 1)
      {
        void *tmp = NULL;
/* (2) */
        size <<= 1;
        tmp = realloc (tab, sizeof (*tab) * size);
        if (tmp != NULL)
        {
          tab = tmp;
        }
        else
        {
          error("str_split Not enough memory");
          Free(tab);
          return NULL;
        }
      }
/* (3) */
      tab[i] = cs;
      s = NULL;
    }
    tab[i] = NULL;
 }
 *nb_tokens= (size_t)i;

 return tab;

} /* str_split */


/* ================== Doxygen Comment ======================================= */
/* ================== Doxygen End Comment =================================== */
char *str_tolower(const char *ct)
{
  char *s = NULL;

  if (ct != NULL)
  {
    int i;

/* (1) */
    Malloc(s, sizeof (*s) * (strlen(ct) + 1));
    if (s != NULL)
    {
/* (2) */
      for (i = 0; ct[i]; i++)
      {
        s[i] = (char)tolower(ct[i]);
      }
      s[i] = '\0';
    }
  }
  return s;

} /* str_tolower */


/* ================== Doxygen Comment ======================================= */
/* ================== Doxygen End Comment =================================== */
char *str_toupper(const char *ct)
{
  char *s = NULL;

  if (ct != NULL)
  {
    int i;

/* (1) */
    Malloc(s, sizeof (*s) * (strlen(ct) + 1));
    if (s != NULL)
    {
/* (2) */
      for (i = 0; ct[i]; i++)
      {
        s[i] = (char)toupper(ct[i]);
      }
      s[i] = '\0';
    }
  }
  return s;

} /* str_toupper */


/* ================== Doxygen Comment ======================================= */
/**
 *  \brief trim_whitespace_noident function removes leading and trailing whitespace.
 *
 *  trim_whitespace_noident function removes leading and trailing whitespace, 
 *  given an char *str string, and return in char * string the filtered
 *  string. It returns as size_t, the length of the string filtered.
 *
 *  Arguments : [INOUT] : char *str
 *
 */
/* ================== Doxygen End Comment =================================== */
char *trim_whitespace_noident(char *str)
{
  char *end = NULL;

  /* Trim leading space */
  while(isspace(*str)) str++;

  /* Trim trailing space */
  end = str + strlen(str) - 1;
  while(end > str && isspace(*end)) end--;

  /* Write new null terminator */
  *(end+1) = 0;

  return str;

} /* trim_whitespace_noident */


/* ================== Doxygen Comment ======================================= */
/**
 *  \brief trim_whitespace_ident function removes leading and trailing whitespace.
 *
 *  trim_whitespace_ident function removes leading and trailing whitespace, 
 *  given an const char *str string, and return in char *out string the filtered
 *  string. It returns as size_t, the length of the string filtered.
 *
 *  Arguments : [OUT] : char *out
 *              [OUT] : size_t len
 *              [IN]  : const char *str
 *
 */
/* ================== Doxygen End Comment =================================== */
size_t trim_whitespace_ident(char *out, size_t len, const char *str)
{
  const char *end;
  size_t out_size;

  /* Trim leading space */
  while(isspace(*str)) str++;

  /* Trim trailing space */
  end = str + strlen(str) - 1;
  while(end > str && isspace(*end)) end--;
  end++;

  /* Set output size to minimum of trimmed string length and buffer size minus 1 */
  out_size = (size_t)(end - str) < len-1 ? (size_t)(end - str) : len-1;

  /* Copy trimmed string and add null terminator */
  memcpy(out, str, out_size);
  out[out_size] = 0;

  return out_size;

} /* trim_whitespace_ident */


/* ================== Doxygen Comment ======================================= */
/* ================== Doxygen End Comment =================================== */
char *ltrim(char *p)
{
  while ( p && isspace(*p) ) ++p ;
  return p ;
}


/* ================== Doxygen Comment ======================================= */
/* ================== Doxygen End Comment =================================== */
char *rtrim(char *p)
{
  char *temp;
  if (!p) return p;
  temp = (char *)(p + strlen(p)-1);
  while ( (temp>=p) && isspace(*temp) ) --temp;
  *(temp+1) = '\0';
  return p;
}


/* ================== Doxygen Comment ======================================= */
/* ================== Doxygen End Comment =================================== */
char *trim(char *p)
{
  return(ltrim(rtrim(p)));
}


/* ================== Doxygen Comment ======================================= */
/* ================== Doxygen End Comment =================================== */
void append_trailing_slash(char *dir)
{
  if (dir == NULL) { return; }
  /* Append trailing "/" character if not present */
  if ( strncmp(&dir[strlen(dir)-1], "/", (size_t)1) != 0 )
  {
    dir = realloc(dir, strlen(dir)+2);
    strncat(dir, "/", (size_t)1);
  }
}


#if 0
/* Uncomment this section below to test module functions */
/* cc com_fct_string.c -o com_fct_string -lm */

/*
int main(void) {


  printf("com_fct_string module functions TEST\n\n");















}*/ /* main */
#endif
