/*---
 *
 * hdf_file
 *
 * Copyright (c) 2007-2012 by
 *
 * Francois Thieuleux (francois dot thieuleux at univ-lille1 dot fr)
 * Fabrice Ducos (fabrice dot ducos at univ-lille1 dot fr)
 * Jerome Riedi (jerome dot riedi at univ-lille1 dot fr)
 *
 * Laboratoire d'Optique Atmospherique
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * ---*/

#include "hdf_file.h"
 
#define PATH_MAX 1024
#define MAX_STRING_LENGTH 65536

char SDS_names[NMAX_SDS][H4_MAX_NC_NAME];           /* Full list names of SDS available for the requested HDF input file */


const char *hdf_format_string(const int32 data_type) /* static buffer inside */
{
  static char format_type[16];

  switch(data_type)
  {
    case DFNT_CHAR    :
    case DFNT_UCHAR   : { strcpy(format_type, "%c"); break ; }
    case DFNT_INT8    : { strcpy(format_type, "%d\t"); break ; }
    case DFNT_UINT8   : { strcpy(format_type, "%u\t"); break ; }
    case DFNT_INT16   : { strcpy(format_type, "%hd\t"); break ; }
    case DFNT_UINT16  : { strcpy(format_type, "%hu\t"); break ; }
    case DFNT_INT32   : { strcpy(format_type, "%ld\t"); break ; }
    case DFNT_UINT32  : { strcpy(format_type, "%lu\t"); break ; }
    case DFNT_FLOAT32 : { strcpy(format_type, "%f\t"); break ; }
    case DFNT_FLOAT64 : { strcpy(format_type, "%lf\t"); break ; }
    default           : fatal("Unknown hdf type code: [%d] %s:%d",(int)data_type,__FILE__,__LINE__);
  }
  return format_type;
}


int get_hdf_file_nb_sds(const char *file)
{
  int32 sd_id;               /**/
  int32 num_datasets;        /**/
  int32 num_global_attrs;    /**/
  intn  status;              /**/

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL) { fatal("Opening read-only file [%s] %s:%d", file,__FILE__,__LINE__);}

  /* Get number of SDS and number of global attributes */
  status = SDfileinfo(sd_id, &num_datasets, &num_global_attrs);

  SDend(sd_id);

  return((int)num_datasets);

} /* get_hdf_file_nb_sds */


int get_hdf_file_nb_attr(const char *file)
{
  int32 sd_id;               /**/
  int32 num_datasets;        /**/
  int32 num_global_attrs;    /**/
  intn  status;              /**/

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL) { fatal("Opening read-only file [%s] %s:%d\n",file,__FILE__,__LINE__); }

  /* Get number of SDS and number of global attributes */
  status = SDfileinfo(sd_id, &num_datasets, &num_global_attrs);

  SDend(sd_id);

  return((int)num_global_attrs);

} /* get_hdf_file_nb_attr */


/*
 *
 *
 */
PTR_SDS* newPTR_SDSs(const char *file, int *sds_index_tab, int nb_sds)
{
  int32 sd_id;
  int32 sds_id;
  char  sds_name[H4_MAX_NC_NAME];
  int32 sds_dimsizes[H4_MAX_VAR_DIMS];
  int32 sds_start[H4_MAX_VAR_DIMS];
  int32 sds_stride[H4_MAX_VAR_DIMS];
  int32 sds_index;
  int32 sds_data_type;
  int32 sds_rank;
  int32 n_attrs;
  intn  status;
  int   size;
  int   isds;                       /* Index over SDS */
  int   j;                          /* Index over rank for a given SDS */

  PTR_SDS *ptr_sds;

  /* Memory allocation */
  Malloc(ptr_sds, (size_t)nb_sds * sizeof *ptr_sds);

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL) { fatal("Opening read-only file [%s] %s:%d", file,__FILE__,__LINE__); }

  /* Loop over SDS */
  for (isds=0;isds<(int)nb_sds;isds++)
  {
    sds_index=(int32)sds_index_tab[isds];

    /* Obtains the data set identifier of the first SDS */
    if ( (sds_id = SDselect(sd_id, sds_index) ) == FAIL) { fatal("Opening SDS [%d] %s:%d",(int)sds_id,__FILE__,__LINE__); }

    /* Retrieves the name, rank, dimension sizes, data type and number of attributes for a data set */
    status = SDgetinfo(sds_id, sds_name, &sds_rank, sds_dimsizes, &sds_data_type, &n_attrs);

    for (j=0;j<(int)sds_rank;j++)
    { 
      sds_start[j]=0;
      sds_stride[j]=1;
    }

    /* Name of each SDS record */
    Malloc(ptr_sds[isds].name, H4_MAX_NC_NAME * sizeof(char));
    strcpy(ptr_sds[isds].name,sds_name);

    /**/
    ptr_sds[isds].type = (int)sds_data_type;
    ptr_sds[isds].rank = (int)sds_rank;

    for (j=0;j<(int)sds_rank;j++) { ptr_sds[isds].dimsizes[j] = (int)sds_dimsizes[j]; }

    /* Size of each SDS record */
    size=1; for (j=0;j<(int)sds_rank;j++) { size*=(int)sds_dimsizes[j]; }
   
    /* Memory allocation for each SDS record */
    switch((int32)sds_data_type)
    {
      case (DFNT_UINT8)  : Malloc(ptr_sds[isds].data, size * sizeof(unsigned char)); break;
      case (DFNT_INT8)   : Malloc(ptr_sds[isds].data, size * sizeof(char)); break;
      case (DFNT_UINT16) : Malloc(ptr_sds[isds].data, size * sizeof(unsigned short int)); break;
      case (DFNT_INT16)  : Malloc(ptr_sds[isds].data, size * sizeof(short int)); break;
      case (DFNT_UINT32) : Malloc(ptr_sds[isds].data, size * sizeof(unsigned int)); break;
      case (DFNT_INT32)  : Malloc(ptr_sds[isds].data, size * sizeof(int)); break;
      case (DFNT_UINT64) : Malloc(ptr_sds[isds].data, size * sizeof(unsigned long int)); break;
      case (DFNT_INT64)  : Malloc(ptr_sds[isds].data, size * sizeof(long int)); break;
      case (DFNT_UINT128): Malloc(ptr_sds[isds].data, size * sizeof(unsigned long long int)); break;
      case (DFNT_INT128) : Malloc(ptr_sds[isds].data, size * sizeof(long long int)); break;
      case (DFNT_FLOAT32): Malloc(ptr_sds[isds].data, size * sizeof(float)); break;
      case (DFNT_FLOAT64): Malloc(ptr_sds[isds].data, size * sizeof(double)); break;
      default            : fatal("Unknown SDS data type %s:%d",__FILE__,__LINE__);
    } /* switch(type) */

    if ( (status = SDreaddata(sds_id, sds_start, sds_stride, sds_dimsizes, ptr_sds[isds].data)) == FAIL)
    {
      free(ptr_sds[isds].data);
      error("Reading SDS [%d] [%s]",(int)status,sds_name);
      return(NULL);
    }

  } /* End loop over required SDS */
  
  assert(ptr_sds);

  return ptr_sds;
} /* newPTR_SDSs */


void freePTR_SDSs(PTR_SDS* ptr_sds, int nb_sds)
{
  for (int i=0;i<nb_sds;i++)
  {
    if (ptr_sds[i].name) free(ptr_sds[i].name);
    if (ptr_sds[i].data) free(ptr_sds[i].data);
  }
  free(ptr_sds);
} /* freePTR_SDSs */


/* Get the SDS index for the specified SDS name */
int get_index_SDS(const char *SDSname, int nb_sds)
{
  char buf[H4_MAX_NC_NAME];
  char *ptrL = NULL;
  char *ptrR = NULL;
  char left_bracket  = '[';                /**/
  char right_bracket = ']';                /**/

  int nb_left_bracket     = 0;             /**/
  int nb_right_bracket    = 0;             /**/
  int index_left_bracket  = 0;             /**/
  int index_right_bracket = 0;             /**/

  ptrL = rindex(SDSname,left_bracket);
  while (ptrL!=NULL)
  {
    nb_left_bracket++;
    index_left_bracket=ptrL-SDSname;
    ptrL=strchr(ptrL+1,left_bracket);
  }

  ptrR = rindex(SDSname,right_bracket);
  while (ptrR!=NULL)
  {
    nb_right_bracket++;
    index_right_bracket=ptrR-SDSname;
    ptrR=strchr(ptrR+1,right_bracket);
  }

#if 0
printf("SDSname:[%s]\n",SDSname);
printf("nb_left_bracket:[%d]\n",nb_left_bracket);
printf("nb_right_bracket:[%d]\n",nb_right_bracket);
printf("index_left_bracket:[%d]\n",index_left_bracket);
printf("index_right_bracket:[%d]\n",index_right_bracket);
#endif

  if ( (nb_left_bracket < nb_right_bracket) ) {         fatal("Missing open square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( (nb_left_bracket > nb_right_bracket) ) {         fatal("Missing close square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( index_left_bracket > index_right_bracket) {      fatal("Close square brackets precede open square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( index_left_bracket == (index_right_bracket-1)) { fatal("Missing array index bounds in square brackets [%s] %s:%d",SDSname,__FILE__,__LINE__); }



  if ( index_left_bracket != 0 ) { memcpy(buf, SDSname, index_left_bracket); buf[index_left_bracket]='\0'; }
  else                           { strncpy(buf, SDSname, H4_MAX_NC_NAME); }

  //if (ptrL!=NULL && ptrR!=NULL) { fprintf(stderr,"aaaa\n"); strncpy(buf, SDSname, index_left_bracket); } 
  //if (ptrL!=NULL && ptrR!=NULL) { strncpy(buf, SDSname, ptrR-SDSname); } 
//  else                          { fprintf(stderr,"bbbb\n"); strncpy(buf, SDSname, H4_MAX_NC_NAME); }

#if 0
  printf("ptrR:[%s]\n",ptrR);
  printf("ptrL:[%s]\n",ptrL);
  printf("buf:[%s]\n",buf);
#endif


  for (int i=0;i<nb_sds;i++)
  {
    //if ((strncmp(SDSname, SDS_names[i], strlen(SDS_names[i])) == 0) && (strlen(SDS_names[i]) == strlen(SDSname)) ) { return i; }
    if ((strncmp(buf, SDS_names[i], strlen(SDS_names[i])) == 0) && (strlen(SDS_names[i]) == strlen(buf)) ) { return i; }
  }
  return -1;

} /* get_index_SDS */


/* Get the array index bounds required for a SDS_name */
/*
 * This function receives an argument which (is supposed) contain a fully SDS name
 * and optionals square brackets with array bounds indexes inside.
 * The array_index[] argument contains valid minimal and maximal array bounds indexes.
 *
 * The function returns 0 when everything is nominal, 1 otherwise.
 *
 * Presence of pair of square brackets is tested, and also array bounds indexes.
 * For example, such a SDS_name like this :
 *                value returned by function    array_index[0] array_index[1]
 * - SDSname[n]                0                     n              n
 * - SDSname[n:]               0                     n          input_value
 * - SDSname[:n]               0                 input_value        n
 * - SDSname[:]                0                 input_value    input_value
 * - SDSname[n:m]              0                     n              m
 * - SDSname                   0                 input_value    input_value
 *
 * When left square bracket number is not equal to right square bracket number,
 * or when square bracket is empty, an error message is returned and execution
 * is aborted.
 *
 */
int get_array_bound_index_SDS(const char *SDSname, int index_array[] )
{
  char *ptr = NULL;                       /**/
  char left_bracket  = '[';               /**/
  char right_bracket = ']';               /**/
  char double_quote = ':';                /**/
  char plus_sign = '+';                /**/
  //char semi_colon=',';                    /**/
  int nb_left_bracket     = 0;            /**/
  int nb_right_bracket    = 0;            /**/
  int index_left_bracket  = 0;            /**/
  int index_right_bracket = 0;            /**/
  char cstr_index[PATH_MAX];              /**/
  size_t cstr_length;                     /**/


  ptr = rindex(SDSname,left_bracket);
  while (ptr!=NULL)
  {
    nb_left_bracket++;
    index_left_bracket=ptr-SDSname;
    ptr=strchr(ptr+1,left_bracket);
  }

  ptr = rindex(SDSname,right_bracket);
  while (ptr!=NULL)
  {
    nb_right_bracket++;
    index_right_bracket=ptr-SDSname;
    ptr=strchr(ptr+1,right_bracket);
  }
#if 0
printf("SDSname:[%s]\n",SDSname);
printf("nb_left_bracket:[%d]\n",nb_left_bracket);
printf("nb_right_bracket:[%d]\n",nb_right_bracket);
#endif
  if ( (nb_left_bracket < nb_right_bracket) ) {         fatal("Missing open square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( (nb_left_bracket > nb_right_bracket) ) {         fatal("Missing close square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( index_left_bracket > index_right_bracket) {      fatal("Close square brackets precede open square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( index_left_bracket == (index_right_bracket-1)) { fatal("Missing array index bounds in square brackets [%s] %s:%d",SDSname,__FILE__,__LINE__); }

  if ( (nb_left_bracket == 0) && (nb_right_bracket == 0) ) { return(0); }                  /* no subset case */

  cstr_length=index_right_bracket-index_left_bracket-1;
  memset(cstr_index,0,sizeof(cstr_index));
  strncpy(cstr_index,&SDSname[index_left_bracket+1],cstr_length);

  if ( (sscanf(cstr_index,"%d%c%d",&index_array[0],&double_quote,&index_array[1]) == 3) || /* [n:m] case */
       (sscanf(cstr_index,"%d%c",&index_array[0],&double_quote) == 2) ) {                  /* [n:]  case */
    if ( strncmp(&double_quote,":",1) == 0 ) { return(0); }
    else { fatal("Array index bounds separator should be "":"" for SDS [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  }
  if (sscanf(cstr_index,"%d",&index_array[0]) == 1) {                                      /* [n] case */
    index_array[1] = index_array[0]; return(0);
  }
  if (sscanf(cstr_index,"%c%d",&double_quote,&index_array[1]) == 2) {                      /* [:m]  case */
    if ( strncmp(&double_quote,":",1) == 0 ) { return(0); }
    else { fatal("Array index bounds separator should be "":"" for SDS [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  }
  if (sscanf(cstr_index,"%c",&double_quote) == 1) {                                        /* [:]   case */
    if ( strncmp(&double_quote,":",1) == 0 ) { return(0); }
    else { fatal("Array index bounds separator should be "":"" for SDS [%c][%s] %s:%d",double_quote,SDSname,__FILE__,__LINE__); }
  }

  fatal("Invalid array index format for SDS [%s] %s:%d",SDSname,__FILE__,__LINE__); /* invalid case */

} /* get_array_bound_index_SDS */


/* Get the array index bounds required for a SDS_name */
/*
 * This function receives an argument which (is supposed) contain a fully SDS name
 * and optionals square brackets with array bounds indexes inside.
 * The array_index[] argument contains valid minimal and maximal array bounds indexes.
 *
 * The function returns 0 when everything is nominal, 1 otherwise.
 *
 * Presence of pair of square brackets is tested, and also array bounds indexes.
 * For example, such a SDS_name like this :
 *                value returned by function    array_index[0] array_index[1]
 * - SDSname[n]                0                     n              n
 * - SDSname[n:]               0                     n          input_value
 * - SDSname[:n]               0                 input_value        n
 * - SDSname[:]                0                 input_value    input_value
 * - SDSname[n:m]              0                     n              m
 * - SDSname                   0                 input_value    input_value
 *
 * When left square bracket number is not equal to right square bracket number,
 * or when square bracket is empty, an error message is returned and execution
 * is aborted.
 *
 */
int get_arrays_bound_index_SDS(const char *SDSname, int index_array[])
{
  char *ptr = NULL;                       /**/

  char left_bracket  = '[';               /**/
  char right_bracket = ']';               /**/

  char double_quote  = ':';               /**/
  char plus_sign     = '+';               /**/
  char semi_colon    = ',';               /**/

  int nb_left_bracket     = 0;            /**/
  int nb_right_bracket    = 0;            /**/
  int index_left_bracket  = 0;            /**/
  int index_right_bracket = 0;            /**/

  char cstr_index[PATH_MAX];              /**/
  size_t cstr_length;                     /**/


  ptr = rindex(SDSname,left_bracket);
  while (ptr!=NULL)
  {
    nb_left_bracket++;
    index_left_bracket=ptr-SDSname;
    ptr=strchr(ptr+1,left_bracket);
  }

  ptr = rindex(SDSname,right_bracket);
  while (ptr!=NULL)
  {
    nb_right_bracket++;
    index_right_bracket=ptr-SDSname;
    ptr=strchr(ptr+1,right_bracket);
  }

  if ( (nb_left_bracket < nb_right_bracket) ) {         fatal("Missing open square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( (nb_left_bracket > nb_right_bracket) ) {         fatal("Missing close square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( index_left_bracket > index_right_bracket) {      fatal("Close square brackets precede open square bracket [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  if ( index_left_bracket == (index_right_bracket-1)) { fatal("Missing array index bounds in square brackets [%s] %s:%d",SDSname,__FILE__,__LINE__); }

  if ( (nb_left_bracket == 0) && (nb_right_bracket == 0) ) { return(0); }                  /* no subset case */

  cstr_length=index_right_bracket-index_left_bracket-1;
  memset(cstr_index,0,sizeof(cstr_index));
  strncpy(cstr_index,&SDSname[index_left_bracket+1],cstr_length);

  if ( (sscanf(cstr_index,"%d%c%d",&index_array[0],&double_quote,&index_array[1]) == 3) || /* [n:m] case */
       (sscanf(cstr_index,"%d%c",&index_array[0],&double_quote) == 2) )                    /* [n:]  case */
  {
    return(BOUNDS_CONTINUE);
  }
  else
  { 
    if (sscanf(cstr_index,"%d%c%d",&index_array[0],&semi_colon,&index_array[1]) == 3)      /* [n,m] case */
    {
      if ( strncmp(&semi_colon,",",1) == 0 )
      {
        return(BOUNDS_DISCONTINUE);
      }
      else { fatal("Array index bounds separator should be double quote ':' or semi colon ',' for SDS [%s] %s:%d",SDSname,__FILE__,__LINE__); }
    }
  }

  if (sscanf(cstr_index,"%d",&index_array[0]) == 1)                                        /* [n] case */
  {
    index_array[1] = index_array[0]; return(BOUNDS_DISCONTINUE);
  }

  if (sscanf(cstr_index,"%c%d",&double_quote,&index_array[1]) == 2)                        /* [:m]  case */
  {
    if ( strncmp(&double_quote,":",1) == 0 ) { return(BOUNDS_CONTINUE); }
    else { fatal("Array index bounds separator should be "":"" for SDS [%s] %s:%d",SDSname,__FILE__,__LINE__); }
  }

  if (sscanf(cstr_index,"%c",&double_quote) == 1)                                          /* [:]   case */
  {
    if ( strncmp(&double_quote,":",1) == 0 ) { return(BOUNDS_CONTINUE); }
  }

  if (sscanf(cstr_index,"%c",&plus_sign) == 1)                                             /* [+]   case */
  {
    if ( strncmp(&plus_sign,"+",1) == 0 ) { return(BOUNDS_NLF); }
  }

  fatal("Invalid array index format for SDS [%s] %s:%d",SDSname,__FILE__,__LINE__); /* invalid case */

} /* get_arrays_bound_index_SDS */


int get_list_SDSs(const char *file)
{
  int32 sd_id;
  int32 sds_id;
  char  sds_name[H4_MAX_NC_NAME];
  int32 sds_dimsizes[H4_MAX_VAR_DIMS];
  int32 sds_index;
  int32 sds_data_type;
  int32 sds_rank;
  int32 n_attrs;
  intn  status;
  int   nb_sds;                       /* Number of SDS in a CALIOP product */
  int   isds;                       /* Index over SDS */

  /* Get number of SDS in an HDF file  */
  nb_sds=get_hdf_file_nb_sds(file);

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL) { fatal("Opening read-only file [%s] %s:%d",file,__FILE__,__LINE__); }

  /* Loop over SDS */
  for (isds=0;isds<(int)nb_sds;isds++)
  {
    sds_index=(int32)isds;

    /* Obtains the data set identifier of the first SDS */
    if ( (sds_id = SDselect(sd_id, sds_index) ) == FAIL) { fatal("Opening SDS [%d] %s:%d",(int)sds_id,__FILE__,__LINE__); }

    /* Retrieves the name, rank, dimension sizes, data type and number of attributes for a data set */
    status = SDgetinfo(sds_id, sds_name, &sds_rank, sds_dimsizes, &sds_data_type, &n_attrs);

    /* Get name of each SDS record */
    strcpy(SDS_names[isds],sds_name);

  } /* for (isds=0;isds<(int)nb_sds;isds++) */
  return 0;
} /* get_list_SDSs */

#define FCF_NBIT_FIELD 7
void display_fcf(const int j, PTR_SDS *sds_requested, const int index_sds, const int language_offset, const unsigned char decode_fcf, const int array_index[],
                char *output_line)
{
  char token[MAX_STRING_LENGTH];

  int msk[FCF_NBIT_FIELD] = {0,0,0,0,0,0,0};
  const unsigned char fcf_field_max[FCF_NBIT_FIELD] = { 7, 3, 3, 3, 7, 1, 5 };

  if ( decode_fcf & 0x00000001 )
  {
    for (int k=0; k<FCF_NBIT_FIELD; k++) { msk[k] = 1; }
  } else {
    if ( decode_fcf & 0x002 ) { msk[0]=1; }
    if ( decode_fcf & 0x004 ) { msk[1]=1; }
    if ( decode_fcf & 0x008 ) { msk[2]=1; }
    if ( decode_fcf & 0x010 ) { msk[3]=1; }
    if ( decode_fcf & 0x020 ) { msk[4]=1; }
    if ( decode_fcf & 0x040 ) { msk[5]=1; }
    if ( decode_fcf & 0x080 ) { msk[6]=1; }
  }

  /* Loop over array elements */
  for (int l=array_index[0];l<=array_index[1];l++)
  {
    int idata=(l-language_offset) + sds_requested[index_sds].dimsizes[1] * j;
    unsigned short int *data = sds_requested[index_sds].data;

    /* Filter over requested bit fields */
    char buf;
    if ( msk[0] == 1 )
    {
      buf = data[idata] & 0x7;
      if ( buf < 0 || buf > fcf_field_max[0] ) { fatal("FeatureClassificationFlags's Feature Type bit field val:%hu (should be more or equal than 0 and less or equal than %d)",buf,fcf_field_max[0]); }
      else                      { sprintf(token, "%hu\t", buf); strcat(output_line, token); }
    }
    if ( msk[1] == 1 )
    {
      buf = data[idata] >> 3 & 0x3;
      if ( buf < 0 || buf > fcf_field_max[1] ) { fatal("FeatureClassificationFlags's Feature Type QA bit field val:%hu (should be more or equal than 0 and less or equal than %u)",buf,fcf_field_max[1]); }
      else                      { sprintf(token, "%hu\t", buf); strcat(output_line, token); }
    }
    if ( msk[2] == 1 )
    {
      buf = data[idata] >> 5 & 0x3;
      if ( buf < 0 || buf > fcf_field_max[2] ) { fatal("FeatureClassificationFlags's Phase bit field val:%hu (should be more or equal than 0 and less or equal than %u)",buf,fcf_field_max[2]); }
      else                      { sprintf(token, "%hu\t", buf); strcat(output_line, token); }
    }
    if ( msk[3] == 1 )
    {
      buf = data[idata] >> 7 & 0x3;
      if ( buf < 0 || buf > fcf_field_max[3] ) { fatal("FeatureClassificationFlags's Phase QA bit field val:%hu (should be more or equal than 0 and less or equal than %u)",buf,fcf_field_max[3]); }
      else                      { sprintf(token, "%hu\t", buf); strcat(output_line, token); }
    }
    if ( msk[4] == 1 )
    {
      buf = data[idata] >> 9 & 0x7;
      if ( buf < 0 || buf > fcf_field_max[4] ) { fatal("FeatureClassificationFlags's Feature Sub Type bit field val:%hu (should be more or equal than 0 and less or equal than %u)",buf,fcf_field_max[4]); }
      else                      { sprintf(token, "%hu\t", buf); strcat(output_line, token); }
    }
    if ( msk[5] == 1 )
    {
      buf = data[idata] >> 12 & 0x1;
      if ( buf < 0 || buf > fcf_field_max[5] ) { fatal("FeatureClassificationFlags's TypeQA bit field val:%hu (should be more or equal than 0 and less or equal than %u)",buf,fcf_field_max[5]); }
      else                      { sprintf(token, "%hu\t", buf); strcat(output_line, token); }
    }
    if ( msk[6] == 1 )
    {
      buf = data[idata] >> 13 & 0x7;
      if ( buf < 0 || buf > fcf_field_max[6] ) { fatal("FeatureClassificationFlags's Horizonthal_Averaging bit field val:%hu (should be more or equal than 0 and less or equal than %u)",buf,fcf_field_max[6]); }
      else                      { sprintf(token, "%hu\t", buf); strcat(output_line, token); }
    }

  } /* for (int l=lmin;l<lmax;l++) */

}
#undef FCF_NBIT_FIELD


int print_list_SDSs(const char *file)
{
  int32 sd_id;                        /* */
  int32 sds_id;                       /* */
  char  sds_name[H4_MAX_NC_NAME];     /* */
  int32 sds_dimsizes[H4_MAX_VAR_DIMS];/* */
  int32 sds_index;                    /* */
  int32 sds_data_type;                /* */
  int32 sds_rank;                     /* */
  int32 n_attrs;                      /* */
  intn  status;                       /* */
  int   nb_sds;                       /* Number of SDS in a CALIOP product */
  int   isds;                         /* Index over SDS */
  int   j;

  char cstr_sds_data_type[8];         /* */

  /* Get number of pixels in a single CALIOP orbit */
  nb_sds=get_hdf_file_nb_sds(file); /**/

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL) { fatal("Opening read-only file [%s] %s:%d",file,__FILE__,__LINE__); }

  /* Loop over SDS */
  for (isds=0;isds<(int)nb_sds;isds++)
  {
    sds_index=(int32)isds;

    /* Obtains the data set identifier of the first SDS */
    if ( (sds_id = SDselect(sd_id, sds_index) ) == FAIL) { fatal("Opening SDS [%d] %s:%d",(int)sds_id,__FILE__,__LINE__); }

    /* Retrieves the name, rank, dimension sizes, data type and number of attributes for a data set */
    status = SDgetinfo(sds_id, sds_name, &sds_rank, sds_dimsizes, &sds_data_type, &n_attrs);

    switch(sds_data_type)
    {
    case (DFNT_UCHAR)  : strcpy(cstr_sds_data_type,"UCHAR");   break;
    case (DFNT_CHAR)   : strcpy(cstr_sds_data_type,"CHAR");    break;
    case (DFNT_UINT8)  : strcpy(cstr_sds_data_type,"UINT8");   break;
    case (DFNT_INT8)   : strcpy(cstr_sds_data_type,"INT8");    break;
    case (DFNT_UINT16) : strcpy(cstr_sds_data_type,"UINT16");  break;
    case (DFNT_INT16)  : strcpy(cstr_sds_data_type,"INT16");   break;
    case (DFNT_UINT32) : strcpy(cstr_sds_data_type,"UINT32");  break;
    case (DFNT_INT32)  : strcpy(cstr_sds_data_type,"INT32");   break;
    case (DFNT_UINT64) : strcpy(cstr_sds_data_type,"UINT64");  break;
    case (DFNT_INT64)  : strcpy(cstr_sds_data_type,"INT64");   break;
    case (DFNT_UINT128): strcpy(cstr_sds_data_type,"UINT128"); break;
    case (DFNT_INT128) : strcpy(cstr_sds_data_type,"INT128");  break;
    case (DFNT_FLOAT32): strcpy(cstr_sds_data_type,"FLOAT32"); break;
    case (DFNT_FLOAT64): strcpy(cstr_sds_data_type,"FLOAT64"); break;
    default            : fatal("Unknown SDS data type %s:%d",__FILE__,__LINE__); break;
    } /* switch(type) */

    fprintf(stdout,"[%2d]:%55s   data_type:%8s   rank:%2d   n_attrs:%2d  ",isds,sds_name,cstr_sds_data_type,(int)sds_rank,(int)n_attrs);
    for (j=0;j<(int)sds_rank;j++) { printf("dimsizes[%2.2d]:%5d  ",j,(int)sds_dimsizes[j]); }
    fprintf(stdout,"\n");

  } /* for (isds=0;isds<(int)nb_sds;isds++) */

  /* Nominal case */
  return 0;
} /* print_list_SDSs */
