/*---
 *
 * polder_geo
 *
 * Copyright (c) 2007-2012 by
 *
 * Francois Thieuleux (francois dot thieuleux at univ-lille1 dot fr)
 * Fabrice Ducos (fabrice dot ducos at univ-lille1 dot fr)
 *
 * Laboratoire d'Optique Atmospherique
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * ---*/

#include "polder_geo.h"

/*
 *
 */
int NINT(double x) { return x < 0.0 ? (int)(x-0.5):(int)(x+0.5); }

/*
 *
 */
int latlon2ligcol_polder(float lat, float lon, int res, float *lig, float *col) {
  int nlines;      /* Absolute lines numbers int the POLDER grid for the requested reolsution res */
  float Ni;       /* Intermediate variables */

  switch (res) {
    case POLDER_FULL_RESOLUTION   : nlines=N_LINES_FULL_RESOLUTION; break; /* Full resolution (1x1) */
    case POLDER_MEDIUM_RESOLUTION : nlines=N_LINES_MEDIUM_RESOLUTION; break; /* Medium resolution (3x3) */
    case POLDER_LOW_RESOLUTION    : nlines=N_LINES_LOW_RESOLUTION; break; /* Low resolution (9x9) */
    default                : fprintf(stderr,"Output POLDER grid resolution not supported\n"); return(1); /* Error case */
  } /* switch */
  
  if ( (lat > +90.) || (lat < -90.) || (lon > +180.) || (lon < -180.) ) {
    fprintf(stderr,"Invalid physical coordinates for the specified (%f,%f) input geographical coordinate couple\n",lat,lon);
    *lig=-99.;
    *col=-99.;
    return(2);
  } /* Test on input validity space domain */

  *lig = (nlines/180.)*(90.-lat)+0.5;
  Ni   = NINT(nlines*sin(M_PI*(*lig-0.5)/nlines));
  *col = nlines+0.5+Ni*lon/180.;

  return 0;	/* Nominal output */
} /* latlon2ligcol_polder */

/*
 *
 */
int ligcol2latlon_polder(float lig, float col, int res, float *lat, float *lon) {
  int nlines;      /* Absolute lines numbers int the POLDER grid for the requested reolsution res */
  float Ni;       /* Intermediate variables */

  switch (res) {
    case POLDER_FULL_RESOLUTION   : nlines=N_LINES_FULL_RESOLUTION; break; /* Full resolution (1x1) */
    case POLDER_MEDIUM_RESOLUTION : nlines=N_LINES_MEDIUM_RESOLUTION; break; /* Medium resolution (3x3) */
    case POLDER_LOW_RESOLUTION    : nlines=N_LINES_LOW_RESOLUTION; break; /* Low resolution (9x9) */
    default                : fprintf(stderr,"Output POLDER grid resolution not supported\n"); return(1); /* Error case */
  } /* switch */

  if ((lig < 0) || (lig > nlines) || (abs(col-0.5-nlines)>NINT(nlines*(sin(M_PI*(lig-0.5)/nlines))))) {
    fprintf(stderr,"Invalid physical coordinates for the specified (%f,%f) input grid coordinate couple\n",lig,col);
    *lat=777.777;
    *lon=777.777;
    return(2);
  } /* Test on input validity space domain */

  *lat = 90.-180./nlines*(lig-0.5);
  Ni   = NINT(nlines*cos(M_PI*(*lat)/180.));
  *lon = 180.*(col-nlines-0.5)/Ni;

  return 0;      /* Nominal output */
} /* ligcol2latlon_polder */


/*
int main() {

int err;

float lat;
float lon;

int res;

float lig;
float col;

lat=87.25;
lon=170.0;

res=1;

printf("INPUT  lat:[%f] lon:[%f]\n",lat,lon);
err=latlon2ligcol_polder(lat,lon,res,&lig,&col);
printf("OUTPUT lig:[%f] col:[%f]\n",lig,col);
printf("err:[%d]\n",err);

printf("INPUT  lig:[%f] col:[%f]\n",lig,col);
err=ligcol2latlon_polder(lig,col,res,&lat,&lon);
printf("OUTPUT lat:[%f] lon:[%f]\n",lat,lon);
printf("err:[%d]\n",err);

lig=1620;
col=3240;

printf("INPUT  lig:[%f] col:[%f]\n",lig,col);
err=ligcol2latlon_polder(lig,col,res,&lat,&lon);
printf("OUTPUT lat:[%f] lon:[%f]\n",lat,lon);
printf("err:[%d]\n",err);

printf("INPUT  lat:[%f] lon:[%f]\n",lat,lon);
err=latlon2ligcol_polder(lat,lon,res,&lig,&col);
printf("OUTPUT lig:[%f] col:[%f]\n",lig,col);
printf("err:[%d]\n",err);


}
*/
