/*---
 *
 * caliop2ascii
 *
 * Copyright (c) 2007-2012 by
 *
 * Francois Thieuleux (francois dot thieuleux at univ-lille1 dot fr)
 * Fabrice Ducos (fabrice dot ducos at univ-lille1 dot fr)
 *
 * Laboratoire d'Optique Atmospherique
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * ---*/

/* */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <strings.h> /* index, rindex */
#include <unistd.h>  /* getopt */
#include <getopt.h>  /* optarg optind optopt */
#include <stdlib.h>  /* setenv */
#include <assert.h>
#include <ctype.h>
#include <libgen.h>
#include <time.h>    /* tzset */

#define _GNU_SOURCE
#include <string.h>  /* memset strstr strcasestr */

/* HDF includes */
#include <hdf.h>
#include <df.h>
#include <dfan.h>
#include <dfgr.h>
#include <dfi.h>
#include <dfrig.h>
#include <dfsd.h>
#include <mfhdf.h>
#include <hcomp.h>
#include <hntdefs.h>
#include <hkit.h>

/* HDF v-datas includes */
#include <vg.h>


#define OUTPUT_FORMAT_MAXLEN 32
#define TIME_FORMAT_MAXLEN 32
#define MAX_STRING_LENGTH 65536

/* Common systems declarations */
#include "system.h"

/* Version number */
#include "version.h"
/* Application name definition */
#include "appname.h"
/* Common structures */
#include "common_struct.h"
/* Debug messages */
#include "debug.h"

#include "verbose.h"

/* General informations about this application (version nb, name, ...) */
#include "caliop2ascii.h"

/* Common errors functions */
#include "com_fct_errors.h"
/* Common strings functions */
#include "com_fct_string.h"
/* Common dates functions */
#include "com_fct_date.h"
/* Common memory allocation functions */
#include "com_fct_xalloc.h"
/* Common POLDER projection functions */
#include "polder_geo.h"

/* Common HDF functions */
#include "hdf_file.h"
 

/* Pointer to the tail component of the filename by which this program was
   invoked.  Used everywhere in error and warning messages. */
const char *program_name = PROGRAM_NAME;
const char *project_name = PROJECT_NAME;

const char *soft_version = SOFT_VERSION;

int verbose = 0;                                        /**/
int debug   = 0;                                        /**/

int allsds  = 0;                                        /**/

static struct option long_options[] =
{
  /* These options set a flag. */
  {"quiet",                        no_argument,       &verbose, 0},
  {"debug",                        no_argument,       &debug, 0},
  /* These options don't set a flag. We distinguish them by their indices. */
  {"all",                          no_argument,       &allsds, 1},
  {"attributes",                   optional_argument, NULL, 'a'},
  {"coordinates_geo",              required_argument, NULL, 'g'},
  {"coordinates_grid",             required_argument, NULL, 'c'},
  {"date",                         no_argument,       NULL, 'd'},
  {"date_format",                  required_argument, NULL, 'D'},
  {"decode_fcf",                   optional_argument, NULL, 'f'},
  {"geo_subset",                   required_argument, NULL, 'G'},
  {"grid_subset",                  required_argument, NULL, 'C'},
  {"help",                         no_argument,       NULL, GETOPT_HELP_CHAR}, 
  {"examples",                     no_argument,       NULL, GETOPT_EXAMPLES_CHAR}, 
  {"index_offset",                 required_argument, NULL, 'i'},
  {"metadata",                     optional_argument, NULL, 'm'},
  {"node_longitude",               no_argument,       NULL, 'N'},
  {"record_index",                 no_argument,       NULL, 'r'},
  {"time",                         no_argument,       NULL, 't'},
  {"time_format",                  required_argument, NULL, 'T'},
  {"verbose",                      optional_argument, NULL, 'v'},
  {"short_hidden_version",         no_argument,       NULL, 'V'},
  {"reverse_coordinates",          no_argument,       NULL, 'X'},
  {"version",                      no_argument,       NULL, GETOPT_VERSION_CHAR},
  {NULL, 0, NULL, 0}
};


//static int time_format_is_set = 0;                                /**/

/*--------------------------------- ----------------------------------- */
/*
  A portable version of GNU timgm, from the man page of timegm
 */
/*--------------------------------- ----------------------------------- */
#ifndef _BSD_SOURCE || _POSIX_C_SOURCE >= 200112L || _XOPEN_SOURCE >= 600
extern int setenv(const char *name, const char *value, int overwrite);
extern int unsetenv(const char *name);
#endif

#ifndef _POSIX_C_SOURCE || _XOPEN_SOURCE
extern void tzset (void);
#endif

time_t my_timegm (struct tm *tm)
{
    time_t ret;
    char *tz;

    tz = getenv("TZ");
    setenv("TZ", "UTC", 1);
    tzset();
    ret = mktime(tm);
    if (tz)
        setenv("TZ", tz, 1);
    else
        unsetenv("TZ");
    tzset();
    return ret;
} /* my_timegm */


/*--------------------------------- ----------------------------------- */
/*
 *   Print synopsis information about application usage.
 *     Display message when bad command use
 */
/*--------------------------------- ----------------------------------- */
static int usage(int status)
{
  if (status != EXIT_SUCCESS)
  {
    fprintf (stderr,"      Try `%s --help' for more information.\n", program_name);
  }
  else
  {
    fprintf(stderr,"      %s version %s\n\n",program_name,soft_version);
    fprintf(stderr,"      Usage : %s [list of options] <input CALIOP HDF file> [SDS_name1,[SDS_name2 ...]]\n\n",program_name);
    fprintf(stderr,"      If invoked without any options and only a supported CALIOP HDF file format (see list above), %s will display the list of available SDS\n",program_name);
    fprintf(stderr,"      If invoked with -all option and only a supported CALIOP HDF file format (see list above), %s will display every available SDS datas\n",program_name);
    fprintf(stderr,"      If invoked without any options and arguments, %s will display this usage message\n",program_name);
    fprintf(stderr,"\n");
    fprintf(stderr,"      Note that for long options, equal sign should be used when passing optional argument value.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      Options\n");
    fprintf(stderr,"      --all                                                display all available SDS\n");
    //fprintf(stderr,"      --attributes|-a                                      display globals (without SDS_name) or specified SDS's attributes\n");
    fprintf(stderr,"      --coordinates_grid|-c [1|2|3]                        display coordinates in the parasol grid (line,column)\n");
    fprintf(stderr,"                                                           optional argument should be\n");
    fprintf(stderr,"                                                                1 for full resolution POLDER's grid (6480x3240),\n");
    fprintf(stderr,"                                                                2 for medium resolution POLDER's grid (2160x1080),\n");
    fprintf(stderr,"                                                                3 for low resolution POLDER's grid (720x360).\n");
    fprintf(stderr,"                                                                (if omitted, full resolution is supposed)\n");
    fprintf(stderr,"      --grid_subset|-C <min_line[:min_column][,max_line[:max_column]]>\n");
    fprintf(stderr,"                                                           display values only in the suitable zone\n");
    fprintf(stderr,"      --date|-d                                            display the acquisition date for each pixel\n");
    fprintf(stderr,"                                                           unless the -T option is used, the default date format is \"%s\"\n",DEFAULT_DATE_FORMAT);
    fprintf(stderr,"      --date_format|-D <date_format>                       sets the format used to display dates\n");
    fprintf(stderr,"                                                           same syntax as the Unix date command, see the man page of date(1)\n");
    //fprintf(stderr,"  --output_format|-F <format>                          output format (in printf-style), default is %s\n", args->output_format);
  //  fprintf(stderr,"  --output_format|-F <format>                          output format (in printf-style), default is \"%-9g\"\n");
    fprintf(stderr,"      --coordinates_geo|-g                                 display geographic coordinates (latitude, longitude)\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      --decode_fcf|-f [fcf_field]                          display fields from specific's Feature_Classification_Flags SDS (if requested) \n");
    fprintf(stderr,"                                                           as mentionned in the Document Data Management System - Data Products Catalog,\n");
    fprintf(stderr,"                                                           No: PC-SCI-503 Release and following releases.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"                                                           Seven fields can be obtained (see Table 44 from previous reference):\n");
    fprintf(stderr,"                                                           Bit(s)   Field Description         \n");
    fprintf(stderr,"                                                           1-3      Feature Type              \n");
    fprintf(stderr,"                                                           4-5      Feature Type QA           \n");
    fprintf(stderr,"                                                           6-7      Ice/Water Phase           \n");
    fprintf(stderr,"                                                           8-9      Ice/Water Phase QA        \n");
    fprintf(stderr,"                                                           10-12    Feature Sub-type          \n");
    fprintf(stderr,"                                                           13       Cloud/Aerosol/PSC Type QA \n");
    fprintf(stderr,"                                                           14-16    Horizonthal averaging required for detection\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"                                                           If no option is given, every fields will be display.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"                                                           Optional argument should be\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"                                                           . FeatureType \n");
    fprintf(stderr,"                                                           . FeaturetypeQA\n");
    fprintf(stderr,"                                                           . Phase\n");
    fprintf(stderr,"                                                           . PhaseQA\n");
    fprintf(stderr,"                                                           . FeatureSubType\n");
    fprintf(stderr,"                                                           . TypeQA\n");
    fprintf(stderr,"                                                           - HA\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"                                                           You can give many times --decode_fcf|-f option with argument,\n");
    fprintf(stderr,"                                                           but --decode_fcf|-f option without argument will take precedence \n");
    fprintf(stderr,"                                                           to others.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      --geo_subset|-G <min_lat[:min_lon],max_lat[:max_lon]>\n");
    fprintf(stderr,"                                                           display values only in the suitable zone\n");
    fprintf(stderr,"      --index_language_offset|-i <c|fortran>               convention to apply for array index offset\n");
    fprintf(stderr,"                                                                  1 for FORTRAN (default convention),\n");
    fprintf(stderr,"                                                                  0 for C\n");
    fprintf(stderr,"      --metadata|-m [metadata_name]                        display metadata list available with their characteristics\n");
    fprintf(stderr,"                                                           Optionally, you can specifiy a metadata metadata_name, and\n");
    fprintf(stderr,"                                                           %s will then display values and then exit.\n",program_name);
    fprintf(stderr,"                                                           (--record_index option is compatible with it).\n");
    fprintf(stderr,"                                                           (For DFNT_CHAR or DFNT_UCHAR metadata type,\n");
    fprintf(stderr,"                                                           strings will be displayed, --record_index option is deactivate).\n");
    fprintf(stderr,"      --node_longitude|-N                                  display the node longitude (between -180 and +180 degrees)\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      --quiet|-q                                           Quiet execution.  Useful if %s is run in a script.\n",program_name);
    fprintf(stderr,"                                                           Note that this option will inhibit --verbose option.\n");
    fprintf(stderr,"      --record_index|-r                                    display the record index as figured in th CALIOP HDF file\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      --time|-t                                            display the acquisition time for each pixel\n");
    fprintf(stderr,"                                                           unless the -T option is used, the default time format is \"%s\"\n", DEFAULT_TIME_FORMAT);
    fprintf(stderr,"      --time_format|-T <time_format>                       sets the format used to display times\n");
    fprintf(stderr,"                                                           same syntax as the Unix date command, see the man page of date(1)\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      --verbose|-v                                         display extended information\n");
    fprintf(stderr,"                                                           This option support level of verbosity as optional argument:\n");
    fprintf(stderr,"                                                           0 (QUIET)      <-- Nothing is reported to sdtout and/or stderr,\n");
    fprintf(stderr,"                                                                              except fatal error messages on stderr that lead to stop execution.\n");
    fprintf(stderr,"                                                           1 (MINIMAL)    <-- Errors messages only are reported to stderr.\n");
    fprintf(stderr,"                                                           2 (NORMAL)     <-- Default value.\n");
    fprintf(stderr,"                                                           3 (DETAILED)   <-- More messages are displayed\n");
    fprintf(stderr,"                                                           4 (DIAGNOSTIC) <-- Debugging purposes (Much more messages displayed\n");
    fprintf(stderr,"                                                           Optional level of verbosity should be given as : --verbose=<level>\n");
    fprintf(stderr,"                                                           By default (-v or --verbose option), the behaviour is same as previously (normal verbose).\n");
    fprintf(stderr,"                                                           Note that level of verbosity could be given as an index (between 0 and 4 included),\n");
    fprintf(stderr,"                                                           or a character string (QUIET, MINIMAL, NORMAL, DETAILED or DIAGNOSTIC, case insensitive).\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      --version                                            display version number and copyright informations then exit\n");
    fprintf(stderr,"      --V                                                  display version number then exit\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      --reverse_coordinates|-X                             display (lines and columns) and (latitudes and longitudes) in reverse order\n");
    fprintf(stderr,"                                                           by default, %s display lines before columns and latitudes before longitudes;\n",program_name);
    fprintf(stderr,"                                                           some tools expect the reverse, which is the -X option aim at.\n");
    fprintf(stderr,"      --examples                                           display some examples and exit\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      --help                                               display this help and examples then exit\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"      CALIOP files types supported by this version\n");
    fprintf(stderr,"         - CAL_LID_L2_333mCLay (Lidar Level 2 Cloud Layer data product at 333 m horizontal resolution)\n");
    fprintf(stderr,"         - CAL_LID_L2_01kmCLay (Lidar Level 2 Cloud Layer data product at 1 km horizontal resolution)\n");
    fprintf(stderr,"         - CAL_LID_L2_05kmCLay (Lidar Level 2 Cloud Layer data product at 5 km horizontal resolution)\n");
    fprintf(stderr,"         - CAL_LID_L2_05kmCPro (Lidar Level 2 Cloud Profile data product at 5 km horizontal resolution)\n");
    fprintf(stderr,"         - CAL_LID_L2_05kmALay (Lidar Level 2 Aerosol Layer data product at 5 km horizontal resolution)\n");
    fprintf(stderr,"         - CAL_LID_L2_VFM (Lidar Level 2 Vertical Feature Mask data product at 5 km horizontal resolution)\n");
    fprintf(stderr,"         - CAL_LID_L1 (Lidar Level 1B Profiles DP 1.1)\n");
    fprintf(stderr,"         - ....\n\n");
    examples(status);

  }

  exit(status);
} /* usage */


/*--------------------------------- ----------------------------------- */
/*
  Display some examples about application usage.
*/
/*--------------------------------- ----------------------------------- */
static int examples(int status)
{
  fprintf(stderr,"---------------------------------- examples ----------------------------------\n\n");
  fprintf(stderr,"      To display the list of available SDSs, and associated informations,\n");
  fprintf(stderr,"      for CALIOP LIDAR Level2 5km cloud products, invoke only with filename as argument :\n");
  fprintf(stderr,"      %s CAL_LID_L2_05kmCLay-Prov-V3-01.2007-08-01T19-04-29ZD.hdf\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display the list of available metadata and associated informations\n");
  fprintf(stderr,"      just invoke as :\n");
  fprintf(stderr,"      %s --metadata CAL_LID_L2_05kmCLay-Prov-V3-01.2007-08-01T19-04-29ZD.hdf\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display values associated to a particular metadata, like Lidar_Data_Altitudes\n");
  fprintf(stderr,"      just invoke as :\n");
  fprintf(stderr,"      %s --metadata=Lidar_Data_Altitudes CAL_LID_L2_05kmCLay-Prov-V3-01.2007-08-01T19-04-29ZD.hdf\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display the value of the (acending or descending) node longitude\n");
  fprintf(stderr,"      just invoke as :\n");
  fprintf(stderr,"      %s -N CAL_LID_L2_05kmCLay-Prov-V3-01.2007-08-01T19-04-29ZD.hdf\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display solar zenith angle, number of layers found, and each of the layer top altitude [km],\n");
  fprintf(stderr,"      with information about grid number line and column (full resolution), geographical coordinates,\n");
  fprintf(stderr,"      date and time for CALIOP LIDAR Level2 5km cloud products, invoke as :\n");
  fprintf(stderr,"      %s -c -g -t -d CAL_LID_L2_05kmCLay-Prov-V2-01.2006-08-23T00-48-35ZD.hdf Solar_Zenith_Angle Number_Layers_Found Layer_Top_Altitude\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display the numbers of layers found, and decoded values of feature classification flags fields,\n");
  fprintf(stderr,"      with information about grid number line and column (full resolution), geographical coordinates,\n");
  fprintf(stderr,"      but only for latitudes between 22.2 and 25.5 degres,\n");
  fprintf(stderr,"      for CALIOP LIDAR Level2 5km cloud products, invoke as :\n");
  fprintf(stderr,"      %s -c -g -f -G 22.2,25.5 CAL_LID_L2_05kmCLay-Prov-V3-01.2007-08-01T19-04-29ZD.hdf Number_Layers_Found Feature_Classification_Flags\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display the numbers of layers found, and top and bottom altitudes\n");
  fprintf(stderr,"      with information about grid number line and column (full resolution), geographical coordinates,\n");
  fprintf(stderr,"      but subsetting to the first three values (over ten),\n");
  fprintf(stderr,"      for CALIOP LIDAR Level2 5km cloud products, invoke as :\n");
  fprintf(stderr,"      %s -c -g CAL_LID_L2_05kmCLay-Prov-V3-01.2007-08-01T19-04-29ZD.hdf Number_Layers_Found Layer_Top_Altitude[1:3] Layer_Bottom_Altitude[1:3]\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      For CALIOP files that have specific Number_Layers_Found's SDS, to display only top and bottom altitudes for numbers of layers found\n");
  fprintf(stderr,"      with information about grid number line and column (full resolution), geographical coordinates,\n");
  fprintf(stderr,"      for CALIOP LIDAR Level2 5km cloud products, invoke as :\n");
  fprintf(stderr,"      %s -c -g CAL_LID_L2_05kmCLay-Prov-V3-01.2007-08-01T19-04-29ZD.hdf Layer_Top_Altitude[+] Layer_Bottom_Altitude[+]\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display extended version informations release\n");
  fprintf(stderr,"      just invoke as :\n");
  fprintf(stderr,"      %s -V\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display version ad copyright informations release\n");
  fprintf(stderr,"      just invoke as :\n");
  fprintf(stderr,"      %s --version\n",program_name);
  fprintf(stderr,"      \n");
  fprintf(stderr,"      To display help and/or usage informations\n");
  fprintf(stderr,"      just invoke as :\n");
  fprintf(stderr,"      %s --help\n",program_name);
  fprintf(stderr,"      \n");

  exit(status);
} /* examples */


/*--------------------------------- ----------------------------------- */
/* Read arguments passed through command line                           */
/*--------------------------------- ----------------------------------- */
static int decodeargs(int *argc, char **argv[], ARGS *args)
{
  char *separator                          = "\t";                 /* Output field separator */

  char output_format[OUTPUT_FORMAT_MAXLEN] = "%-9g";               /**/
  char geodetic_coord_format[OUTPUT_FORMAT_MAXLEN] = "%-9g";       /**/

  char time_format[TIME_FORMAT_MAXLEN] = "";                       /**/
  char date_format[TIME_FORMAT_MAXLEN] = "";                       /**/

  static int quiet;                                                /**/

  static int attributes = 0;                                       /**/
  char       attributes_fieldname[PATH_MAX];                       /**/
  static int attributes_field = 0;                                 /**/

  static int coord_polder     = 0;                                 /**/
  static int coord_latlon     = 0;                                 /**/
  static int coord_time       = 0;                                 /**/
  static int coord_date       = 0;                                 /**/

  static unsigned char decode_fcf = 0;                                 /**/
  char       fcf_fieldname[PATH_MAX];                              /**/

  static int node_longitude = 0;                                   /**/

  static int record_index = 0;                                     /**/

  static int reverse_coord    = 0;                                 /**/

  static int metadata = 0;                                         /**/
  char       metadata_fieldname[PATH_MAX];                         /**/
  static int metadata_field = 0;                                   /**/

  static int resolution       = POLDER_FULL_RESOLUTION;            /**/

  static int language_offset  = 1;                                 /* Fortran array index offset */

  int min_line         = 0;                                        /* Minimal line index in the POLDER grid specified by user used to filter the display */
  int min_column       = 0;                                        /* Minimal column index in the POLDER grid specified by user used to filter the display */
  int max_line         = N_LINES_FULL_RESOLUTION;                  /* Maximal line index in the POLDER grid specified by user used to filter the display */
  int max_column       = N_COLUMNS_FULL_RESOLUTION;                /* Maximal column index in the POLDER grid specified by user used to filter the display */

  float min_latitude   = -90.;                                     /* Minimal geographical latitude value specified by user used to filter the display */
  float max_latitude   = +90.;                                     /* Maximal geographical latitude value specified by user used to filter the display */
  float min_longitude  = -180.;                                    /* Minimal geographical longitude value specified by user used to filter the display */
  float max_longitude  = +180.;                                    /* Maximal geographical longitude value specified by user used to filter the display */

  int time_format_is_set = 0;                                      /**/
  int date_format_is_set = 0;                                      /**/

  int errors = 0;                                                  /* Errors counter */

  int option_index = 0;

  int optc;

  program_name = strrchr (*argv[0], DIR_SEPARATOR);
# ifdef DIR_SEPARATOR_2
  {
    char *slash;

    slash = strrchr (program_name ? program_name : *argv[0], DIR_SEPARATOR_2);
    if (slash)
      program_name = slash;
  }
# endif
  program_name = program_name ? program_name+1 : *argv[0];

  /* getopt_long stores the option index here. */
  while ((optc = getopt_long_only(*argc, *argv, "ac::C:dD:f:F:gG:i:m::NqtT:vVrX", long_options, &option_index)) != -1)
  {
    switch (optc)
    {
      case 0: /* If this option set a flag, do nothing else now. */
        if (long_options[option_index].flag != 0) { break; }
        printf ("option %s", long_options[option_index].name);
        if (optarg) { printf (" with arg %s", optarg); }
        printf ("\n");
        break;
      case 'a' : /* Display global or SD attributes */
        memset(attributes_fieldname,'\0',PATH_MAX);
        if (optarg) { sscanf(optarg,"%s",attributes_fieldname); attributes_field++; }
        attributes++;
        break;
      case 'c' : /* Output POLDER grid coordinates */
        if ( optarg && *optarg ) { resolution = atoi(optarg); }
        if ( (resolution < POLDER_FULL_RESOLUTION) || (resolution > POLDER_LOW_RESOLUTION)) { error("Invalid value for option -c"); errors++; }
        switch(resolution)
        {
          case (POLDER_FULL_RESOLUTION) :
            max_line=N_LINES_FULL_RESOLUTION;
            max_column=N_COLUMNS_FULL_RESOLUTION;
            break;
          case (POLDER_MEDIUM_RESOLUTION) :
            max_line=N_LINES_MEDIUM_RESOLUTION;
            max_column=N_COLUMNS_MEDIUM_RESOLUTION;
            break;
          case (POLDER_LOW_RESOLUTION) :
            max_line=N_LINES_LOW_RESOLUTION;
            max_column=N_COLUMNS_LOW_RESOLUTION;
            break;
          default :
            assert(0);
            break; 
        } /* switch(resolution) */
        coord_polder++;
        break;
      case 'C' : /* Filter output lines under grid coordinates */
        if (sscanf(optarg,"%d:%d,%d:%d",&min_line,&min_column,&max_line,&max_column) == 4 )
        {
          if ( min_line   > max_line )   { error("Inconsistency in argument for option -C (min_line > max_line)"); errors++; }
          if ( min_column > max_column ) { error("Inconsistency in argument for option -C (min_column > max_column)"); errors++; }
        }
        else if (sscanf(optarg,"%d:%d",&min_line,&min_column) == 2)
        {
          max_line = min_line;
          max_column = min_column;
        }
        else if (sscanf(optarg,"%d,%d",&min_line,&max_line) == 2 )
        {
          if ( min_line > max_line ) { error("Inconsistency in argument for option -C (min_line > max_line)"); errors++; }
        }
        else { error("invalid argument for option -C (min_line[:min_column][,max_line[:max_column]])"); errors++; }
        break;
      case 'd' :
        if (! date_format_is_set)
        {
          if (date_format[0] != '\0') { strcat(date_format, "\t"); }
          strcat(date_format, DEFAULT_DATE_FORMAT);
        }
        coord_date++;
        break;
      case 'D' :
        memset(date_format, 0, sizeof(date_format));
        strncpy(date_format, optarg, sizeof(date_format) - 1);
        date_format_is_set = 1;
        break;
      case 'F' : /* output format */
        if ((strlen(optarg) > 2) && (optarg[0] == 'g') && (optarg[1] == ':')) {
          strncpy(geodetic_coord_format, &optarg[2], sizeof(geodetic_coord_format) - 1);
        }
        else {
          snprintf(output_format, OUTPUT_FORMAT_MAXLEN, "%s\t", optarg);
        }
        break;
      case 'f' : /* Feature_Classification_Flags decoding requested */
        if (optarg && *optarg)
        {
          char buf[20];
          if (sscanf(optarg,"%s",buf) == 1)
          {
	    if       ( (strncasecmp(optarg,"FeatureTypeQA",13) == 0) )  { decode_fcf+=4; break; }
	    else if  ( (strncasecmp(optarg,"FeatureType",11) == 0) )    { decode_fcf+=2; break; }
            else if  ( (strncasecmp(optarg,"PhaseQA",7) == 0) )         { decode_fcf+=16; break; }
            else if  ( (strncasecmp(optarg,"Phase",5) == 0) )           { decode_fcf+=8; break; }
            else if  ( (strncasecmp(optarg,"FeatureSubType",14) == 0) ) { decode_fcf+=32; break; }
            else if  ( (strncasecmp(optarg,"TypeQA",6) == 0) )          { decode_fcf+=64; break;  }
            else if  ( (strncasecmp(optarg,"HA",2) == 0) )              { decode_fcf+=128; break; }
          }
          else
          {
            error("Inconsistency in --decode_fcf|-f option argument");
            usage(EXIT_FAILURE);
            return(EXIT_FAILURE);
          }
        } /* if (optarg) */
        decode_fcf++;
        break;
      case 'g' : /* Output geodetic coordinates */
        coord_latlon++;
        break;
      case 'G' : /* Filter output lines under geographical coordinates */
        if (sscanf(optarg,"%f:%f,%f:%f",&min_latitude,&min_longitude,&max_latitude,&max_longitude) == 4 )
        {
          if ( min_latitude  > max_latitude )  { error("Inconsistency in argument for option -G (min_latitude > max_latitude)"); errors++; }
          if ( min_longitude > max_longitude ) { error("Inconsistency in argument for option -G (min_longitude > max_longitude)"); errors++; }
        }
        else if (sscanf(optarg,"%f,%f",&min_latitude,&max_latitude) == 2 )
        {
          if ( min_latitude > max_latitude ) { error("Inconsistency in argument for option -G (min_latitude > max_latitude)"); errors++; }
        }
        else { error("Invalid argument for option -G (min_latitude[:min_longitude],max_latitude[:max_longitude])"); errors++; }
        break;
      case 'i' : /* Array index offset convention */
        if      ( (strncasecmp(optarg,"c",1) == 0) ) { language_offset=C_OFFSET; }
        else if ( (strncasecmp(optarg,"Fortran",7) == 0) || (strncasecmp(optarg,"F",1) == 0) ) { language_offset=FORTRAN_OFFSET; }
        else { 
          error("Invalid value for option -i (should be F/fortran or C/c)\n\n");
          usage(EXIT_FAILURE);
        }
        break;
      case 'm' : /* Display only Lidar Metadata Record  */
        memset(metadata_fieldname,'\0',PATH_MAX);
        if (optarg) { sscanf(optarg,"%s",metadata_fieldname); metadata_field++; }
        metadata++;
        break;
      case 'q' : /* Display only Lidar Metadata Record  */
        quiet++;
        break;
      case 'N' :
        node_longitude++;
        if (optarg) { printf (" with arg %s", optarg); }
        break;
      case 'r' :
        record_index++;
        break;
      case 'X' :
        reverse_coord++;
        break;
      case 't' :
        if (! time_format_is_set)
        {
          if (time_format[0] != '\0') { strcat(time_format, "\t"); }
          strcat(time_format, DEFAULT_TIME_FORMAT);
        }
        coord_time++;
        break;
      case 'T' :
        memset(time_format, 0, sizeof(time_format));
        strncpy(time_format, optarg, sizeof(time_format) - 1);
        time_format_is_set = 1;
        break;
      case 'v':
        verbose = NORMAL_VERBOSE;
        if (optarg)
        {
          char buf[20];
          if (sscanf(optarg,"%d",&verbose) == 1 )
          {
            if ((verbose < QUIET_VERBOSE) || (verbose > DIAGNOSTIC_VERBOSE))
            {
              error("Inconsistency in -v value option argument; Value should be less than %d (DIAGNOSTIC_VERBOSE) and more than %d (QUIET_VERBOSE)",
                    DIAGNOSTIC_VERBOSE,QUIET_VERBOSE);
              usage(EXIT_FAILURE);
              return(EXIT_FAILURE);
            }
          }
          else if (sscanf(optarg,"%s",buf) == 1)
          {
            if      (!strcmp(str_toupper(buf), "QUIET"))      { verbose = QUIET_VERBOSE; }
            else if (!strcmp(str_toupper(buf), "MINIMAL"))    { verbose = MINIMAL_VERBOSE; }
            else if (!strcmp(str_toupper(buf), "NORMAL"))     { verbose = NORMAL_VERBOSE; }
            else if (!strcmp(str_toupper(buf), "DETAILED"))   { verbose = DETAILED_VERBOSE; }
            else if (!strcmp(str_toupper(buf), "DIAGNOSTIC")) { verbose = DIAGNOSTIC_VERBOSE; }
            else
            {
              error("Inconsistency in -v option argument; String should be one of these: QUIET, MINIMAL, NORMAL, DETAILED or DIAGNOSTIC");
              usage(EXIT_FAILURE);
              return(EXIT_FAILURE);
            }
          }
          else
          {
            error("Inconsistency in -v option argument");
            usage(EXIT_FAILURE);
            return(EXIT_FAILURE);
          }
        } /* if (optarg) */
        break;
      case 'V': /* Print version number */
        printf("%s\n",soft_version);
        exit(EXIT_SUCCESS);
        break;
      case GETOPT_EXAMPLES_CHAR:
        printf("%s version %s\n",program_name,soft_version);
        examples(EXIT_SUCCESS);
        break;
      case GETOPT_HELP_CHAR:
        usage(EXIT_SUCCESS);
        break;
      case GETOPT_VERSION_CHAR:
        version_etc (stdout, program_name, project_name, soft_version, AUTHORS, (char *) NULL);
        exit(EXIT_SUCCESS); 
        break;
      case '?':
        /* getopt_long already printed an error message. */
        
        if (isprint (optopt)) fprintf (stderr, "unknown option `-%c'.\n",optopt);
        else fprintf (stderr, "unknown option character `\\x%x'.\n",optopt);
        exit(7);

        break;
      case ':' :
        fatal("Option -%c requires an operand %s:%d",optopt,__FILE__,__LINE__);
      default:
        abort ();

    } /* switch (c)*/
  } /* while (1) */

  *argc -= optind - 1;
  *argv += optind - 1;

  /* Degraded case where errors are encountered */
  if (errors) { usage(EXIT_FAILURE); }
 
  /* Print synospis information when no argument is given */
  //if (optind == 1) { usage(EXIT_FAILURE); }

  /**/
  args->verbose = verbose;
  args->quiet = quiet;
  args->debug = debug;

  args->allsds = allsds;

  args->attributes = attributes;

  args->coord_date = coord_date;

  args->coord_time = coord_time;

  args->time_format_is_set = time_format_is_set;
  args->date_format_is_set = date_format_is_set;

  sprintf(args->time_format,"%s%s",time_format, separator); /**/
  sprintf(args->date_format,"%s%s",date_format, separator); /**/

  args->metadata = metadata;
  if (metadata_fieldname != NULL)
  {
    strcpy(args->metadata_fieldname,metadata_fieldname);
    args->metadata_field = metadata_field;
  }

  args->language_offset = language_offset;

  args->node_longitude = node_longitude;

  args->record_index = record_index;

  args->resolution = resolution;

  args->coord_polder = coord_polder;
  args->coord_latlon = coord_latlon;

  /* Display all Feature_Classification_Flags's specific SDS if any --decode_fcf|-f option without argument is specified */
  if ( decode_fcf & 0x00000001 ) { args->decode_fcf = 1; }
  else                           { args->decode_fcf = decode_fcf; }

  args->min_latitude = min_latitude;
  args->max_latitude = max_latitude;
  args->min_longitude = min_longitude;
  args->max_longitude = max_longitude;

  args->min_line = min_line;
  args->max_line = max_line;
  args->min_column = min_column;
  args->max_column = max_column;

  sprintf(args->output_format,"%s%s",output_format, separator); /**/

  return(EXIT_SUCCESS);

} /* decodeargs */


/**/
int get_caliop_file_nb_record(const char *file)
{
  int32 sd_id;
  int32 dim_sizes[H4_MAX_VAR_DIMS];
  int32 sds_id;
  int32 data_type;
  int32 n_attrs;
  int32 rank;

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL ) { fatal("Unable to open file in read-only mode [%s] %s:%d",file,__FILE__,__LINE__); }

  /* Obtains the data set identifier of the first SDS */
  if ( (sds_id = SDselect(sd_id, 0)) == FAIL ) { fatal("SDselect sd_id:[%d] %s:%d",(int)sd_id,__FILE__,__LINE__); }

  /* Retrieves the name, rank, dimension sizes, data type and number of attributes for a data set */
  if ( SDgetinfo(sds_id, NULL, &rank, dim_sizes, &data_type, &n_attrs) == FAIL ) { fatal("SDgetinfo sds_id:[%d] %s:%d",(int)sds_id,__FILE__,__LINE__); }

  SDendaccess(sds_id);
  SDend(sd_id);

  return(dim_sizes[0]);
} /* get_caliop_file_nb_record */



/*
 *
 *
 */
double get_caliop_file_first_time(const char *file)
{
  int32 sd_id;
  int32 tim_id;
  int32 sds_dimsizes[2];
  int32 sds_start[2]={0,0};
  int32 *sds_stride=NULL;
  int32 sds_data_type;
  int32 sds_rank;
  int32 n_attrs;
  intn  status;

  void *tim_buf=NULL;

  double *tim=NULL;

  double first_tim;

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL ) { fatal("Unable to open file in read-only mode [%s] %s:%d",file,__FILE__,__LINE__); }

  if ( (tim_id = SDselect(sd_id, SDnametoindex(sd_id, "Profile_Time")) ) == FAIL ) { fatal("SDselect Profile_Time sd_id:[%d] %s:%d",(int)tim_id,__FILE__,__LINE__); }

  status = SDgetinfo(tim_id, NULL, &sds_rank, sds_dimsizes, &sds_data_type, &n_attrs);

  /* Size of each SDS record */
  int size=1; for (int j=0;j<(int)sds_rank;j++) { size*=(int)sds_dimsizes[j]; }

  Malloc(tim_buf, size * sizeof *tim_buf);
  if ( (status = SDreaddata(tim_id, sds_start, sds_stride, sds_dimsizes, tim_buf)) == FAIL)
  {
    free(tim);
    fatal("Reading SDS Profile_Time status:[%d] %s:%d",(int)status,__FILE__,__LINE__);
  }
  tim=(double*)tim_buf;

  first_tim = tim[0];

  free(tim);

  return first_tim;
} /* get_caliop_file_first_time */


double get_caliop_file_last_time(const char *file)
{
  int32 sd_id;
  int32 tim_id;
  int32 sds_dimsizes[2];
  int32 sds_start[2]={0,0};
  int32 *sds_stride=NULL;
  int32 sds_data_type;
  int32 sds_rank;
  int32 n_attrs;
  intn  status;

  void *tim_buf=NULL;

  double *tim=NULL;

  double last_tim;

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL ) { fatal("Unable to open file in read-only mode [%s] %s:%d",file,__FILE__,__LINE__); }

  if ( (tim_id = SDselect(sd_id, SDnametoindex(sd_id, "Profile_Time")) ) == FAIL ) { fatal("SDselect Profile_Time sd_id:[%d] %s:%d",(int)tim_id,__FILE__,__LINE__); }

  status = SDgetinfo(tim_id, NULL, &sds_rank, sds_dimsizes, &sds_data_type, &n_attrs);

  /* Size of each SDS record */
  int size=1; for (int j=0;j<(int)sds_rank;j++) { size*=(int)sds_dimsizes[j]; }

  Malloc(tim_buf, size * sizeof *tim_buf);
  if ( (status = SDreaddata(tim_id, sds_start, sds_stride, sds_dimsizes, tim_buf)) == FAIL)
  {
    free(tim);
    fatal("Reading SDS Profile_Time status:[%d] %s:%d",(int)status,__FILE__,__LINE__);
  }
  tim=(double*)tim_buf;

  last_tim = tim[(int)sds_dimsizes[0]-1];


  free(tim);

  return last_tim;
} /* get_caliop_file_last_time */


/*
 *
 *
 */
double get_caliop_file_node_time(const char *file)
{
  int32 sd_id;
  int32 latitude_id,longitude_id,tim_id;
  int32 sds_dimsizes[2];
  int32 sds_start[2]={0,0};
  int32 *sds_stride=NULL;
  int32 sds_data_type;
  int32 sds_rank;
  int32 n_attrs;
  intn  status;
  int   fst_sign=0;
  int   scd_sign=0;
  int   nb_rec;

  void *latitude_buf=NULL;
  void *longitude_buf=NULL;
  void *tim_buf=NULL;

  float *latitude=NULL;
  float *longitude=NULL;
  double *tim=NULL;

  double node_tim;

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL) { fatal("Opening read-only file [%s] %s:%d",file,__FILE__,__LINE__); }

  /* Test if SDS latitude and longitude exists */
  if ( (latitude_id = SDselect(sd_id, SDnametoindex(sd_id, "Latitude")) ) == FAIL)   { fatal("While opening SDS Latitude %s:%d",__FILE__,__LINE__); }
  if ( (longitude_id = SDselect(sd_id, SDnametoindex(sd_id, "Longitude")) ) == FAIL) { fatal("While opening SDS Longitude %s:%d",__FILE__,__LINE__); }

  if ( (tim_id = SDselect(sd_id, SDnametoindex(sd_id, "Profile_Time")) ) == FAIL)          { fatal("While opening SDS Profile_Time %s:%d",__FILE__,__LINE__); }

  status = SDgetinfo(latitude_id, NULL, &sds_rank, sds_dimsizes, &sds_data_type, &n_attrs);

  /* Size of each SDS record */
  int size=1; for (int j=0;j<(int)sds_rank;j++) { size*=(int)sds_dimsizes[j]; }
  nb_rec = (int)sds_dimsizes[0];

  Malloc(latitude_buf, size * sizeof *latitude_buf);
  if ( (status = SDreaddata(latitude_id, sds_start, sds_stride, sds_dimsizes, latitude_buf)) == FAIL)
  {
    fatal("Reading SDS Latitude [%d] %s:%d",(int)status,__FILE__,__LINE__);
  }
  latitude=latitude_buf;

  Malloc(longitude_buf, size * sizeof *longitude_buf);
  if ( (status = SDreaddata(longitude_id, sds_start, sds_stride, sds_dimsizes, longitude_buf)) == FAIL)
  {
    fatal("Reading SDS Longitude [%d] %s:%d",(int)status,__FILE__,__LINE__);
  }
  longitude=longitude_buf;

  Malloc(tim_buf, size * sizeof *tim_buf);
  if ( (status = SDreaddata(tim_id, sds_start, sds_stride, sds_dimsizes, tim_buf)) == FAIL)
  {
    fatal("Reading SDS Profile_Time [%d] %s:%d",(int)status,__FILE__,__LINE__);
  }
  tim=tim_buf;

  /* Get the index for when the sign of the latitude change */
  fst_sign = +1;
  if ( latitude[0] < 0. ) { fst_sign = -1; }

  scd_sign=fst_sign;
  int j=0;
  int i=1;
 
  while ((scd_sign == fst_sign) && (i < nb_rec))
  {
    j = 1 + sds_dimsizes[1] * i;

    scd_sign = +1;
    if ( latitude[j] < 0. ) { scd_sign = -1; }

    i++;

  }
  free(latitude);
  free(longitude);

  node_tim = (tim[j-sds_dimsizes[1]] + tim[j]) / 2.;

  free(tim);

  return node_tim;
} /* get_caliop_file_node_time */

/*
 *
 *
 */
float get_caliop_file_node_longitude(const char *file, int flag_start_stop)
{
  int32 sd_id;
  int32 sds_dimsizes[2];
  int32 sds_start[2]={0,0};
  int32 *sds_stride=NULL;
  int32 sds_data_type;
  int32 sds_rank;
  int32 n_attrs;
  intn  status;
  int   j,jj;
  int   fst_sign=0;
  int   scd_sign=0;
  int   size;
  int   nb_rec;

  float *latitude_sds=NULL;
  float *longitude_sds=NULL;

  float latitude;
  float *latitude_start=NULL;
  float *latitude_stop=NULL;

  float longitude;
  float previous_longitude;
  float *longitude_start=NULL;
  float *longitude_stop=NULL;

  int lat_index = -1;
  int lon_index = -1;

  int lon_start_index = -1;
  int lon_stop_index = -1;

  int lat_start_index = -1;
  int lat_stop_index = -1;

  float node_longitude = -999.;

  /* Open HDF data file in read-only mode */
  if ( (sd_id = SDstart(file, DFACC_READ)) == FAIL) { fatal("Opening read-only file [%s] %s:%d",file,__FILE__,__LINE__); }

  if ( flag_start_stop == FALSE )
  {
    if ( (lat_index = SDselect(sd_id, SDnametoindex(sd_id, "Latitude")) )  == FAIL) { fatal("Opening SDS Latitude %s:%d",__FILE__,__LINE__); }
    if ( (lon_index = SDselect(sd_id, SDnametoindex(sd_id, "Longitude")) ) == FAIL) { fatal("Opening SDS Latitude %s:%d",__FILE__,__LINE__); }

    status = SDgetinfo(lat_index, NULL, &sds_rank, sds_dimsizes, &sds_data_type, &n_attrs);

    size   = (int)sds_dimsizes[0]*(int)sds_dimsizes[1];
    nb_rec = (int)sds_dimsizes[0];

    Malloc(latitude_sds, size * sizeof *latitude_sds);
    if ( (status = SDreaddata(lat_index, sds_start, sds_stride, sds_dimsizes, latitude_sds)) == FAIL) { fatal("Reading SDS Latitude [%d] %s:%d",(int)status,__FILE__,__LINE__); }

    Malloc(longitude_sds, size * sizeof *longitude_sds);
    if ( (status = SDreaddata(lon_index, sds_start, sds_stride, sds_dimsizes, longitude_sds)) == FAIL) { fatal("Reading SDS Longitude [%d] %s:%d",(int)status,__FILE__,__LINE__); }

  } else {

    if ( (lat_start_index = SDselect(sd_id, SDnametoindex(sd_id, "Latitude_Start")) ) == FAIL) { fatal("Opening SDS Latitude_Start %s:%d",__FILE__,__LINE__); }
    if ( (lon_start_index = SDselect(sd_id, SDnametoindex(sd_id, "Longitude_Start")) ) == FAIL) { fatal("Opening SDS Longitude_Start %s:%d",__FILE__,__LINE__); }
    if ( (lat_stop_index = SDselect(sd_id, SDnametoindex(sd_id, "Latitude_Stop")) )   == FAIL) { fatal("Opening SDS Latitude_Stop %s:%d",__FILE__,__LINE__); }
    if ( (lon_stop_index = SDselect(sd_id, SDnametoindex(sd_id, "Longitude_Stop")) )  == FAIL) { fatal("Opening SDS Longitude_Stop %s:%d",__FILE__,__LINE__); }

    status = SDgetinfo(lat_start_index, NULL, &sds_rank, sds_dimsizes, &sds_data_type, &n_attrs);

    size   = (int)sds_dimsizes[0]*(int)sds_dimsizes[1];
    nb_rec = (int)sds_dimsizes[0];

    Malloc(latitude_start, size * sizeof *latitude_start);
    if ( (status = SDreaddata(lat_start_index, sds_start, sds_stride, sds_dimsizes, latitude_start)) == FAIL) { fatal("Reading SDS Latitude [%d] %s:%d",(int)status,__FILE__,__LINE__); }

    Malloc(longitude_start, size * sizeof *longitude_start);
    if ( (status = SDreaddata(lon_start_index, sds_start, sds_stride, sds_dimsizes, longitude_start)) == FAIL) { fatal("Reading SDS Longitude [%d] %s:%d",(int)status,__FILE__,__LINE__); }

    Malloc(latitude_stop, size * sizeof *latitude_stop);
    if ( (status = SDreaddata(lat_stop_index, sds_start, sds_stride, sds_dimsizes, latitude_stop)) == FAIL) { fatal("Reading SDS Latitude [%d] %s:%d",(int)status,__FILE__,__LINE__); }

    Malloc(longitude_stop, size * sizeof *longitude_stop);
    if ( (status = SDreaddata(lon_stop_index, sds_start, sds_stride, sds_dimsizes, longitude_stop)) == FAIL) { fatal("Reading SDS Longitude [%d] %s:%d",(int)status,__FILE__,__LINE__); }
  }

  jj=0;
  j=1;

  if (sds_dimsizes[1] > 1) jj = 1 + sds_dimsizes[1] * j;

  if ( flag_start_stop == TRUE )
  {
    latitude  = ( latitude_start[jj]  + latitude_stop[jj]  ) / 2.;
    longitude = ( longitude_start[jj] + longitude_stop[jj] ) / 2.;
  } else {
    latitude  = latitude_sds[jj];
    longitude = longitude_sds[jj];
  }

  /* Get the index for when the sign of the latitude change */
  fst_sign = +1;
  if ( latitude < 0. ) { fst_sign = -1; }

  scd_sign=fst_sign;
 
  while ((scd_sign == fst_sign) && (j < nb_rec))
  {
    jj = j;
    if (sds_dimsizes[1] > 1) jj = 1 + sds_dimsizes[1] * j;

    if ( flag_start_stop == TRUE )
    {
      latitude  = ( latitude_start[jj]  + latitude_stop[jj]  ) / 2.;
      longitude = ( longitude_start[jj] + longitude_stop[jj] ) / 2.;
    } else {
      latitude  = latitude_sds[jj];
      longitude = longitude_sds[jj];
    }

    scd_sign = +1;
    if ( latitude < 0. ) { scd_sign = -1; }

    j++;

    previous_longitude = longitude;
  }
  jj++;

  if ( flag_start_stop == TRUE )
  {
    latitude  = ( latitude_start[jj]  + latitude_stop[jj]  ) / 2.;
    longitude = ( longitude_start[jj] + longitude_stop[jj] ) / 2.;
  } else {
    latitude  = latitude_sds[jj];
    longitude = longitude_sds[jj];
  }

  node_longitude = (previous_longitude + longitude) / 2.;

  if ( flag_start_stop == TRUE ) {
    free(latitude_start); free(longitude_start);
    free(latitude_stop); free(longitude_stop);
  } else {
    free(latitude_sds); free(longitude_sds);
  }

  return node_longitude;

} /* get_caliop_file_node_longitude */


/*
 *
 */
int main(int argc, char *argv[])
{
  char filein[PATH_MAX];                                  /* */

  int nb_rec=0;                                           /* Number of pixels */
  int nb_sds=0;                                           /* Number total of SDS for the current HDF file */

  int nb_sds_requested=0;                                 /* Total number of requested SDS (excluding those implied by flag option (ex. -g, -t )) */
  int nb_sds_required=0;                                  /* Total number of mandatory SDS (including those implied by flag option (ex. -g, -t)) */

  int j;                                                  /* */
  int k;
  int lig;                                                /* Index of line in the POLDER grid for requested resolution */
  int col;                                                /* Index of line in the POLDER grid for requested resolution */
   
  char sds_name_required_tab[NMAX_SDS][H4_MAX_NC_NAME];      /* */
  char sds_name_requested_tab[NMAX_SDS][H4_MAX_NC_NAME];     /* */
  char sds_name[H4_MAX_NC_NAME];                             /* */

  char token[MAX_STRING_LENGTH];                          /* Output line intermediate buffer */
  char output_line[MAX_STRING_LENGTH];                    /* Output line buffer */

  char cstr_first_time[TIME_FORMAT_MAXLEN + 1];            /* Current formatted first time [metadata] to display */
  char cstr_last_time[TIME_FORMAT_MAXLEN + 1];             /* Current formatted last time [metadata] to display */
  char cstr_node_time[TIME_FORMAT_MAXLEN + 1];             /* Current formatted node longitude time [estimated] to display */
  char cstr_current_time[TIME_FORMAT_MAXLEN + 1];          /* Current formatted time to display */

//  char first_date[DATE_LENGTH + 1];                       /* Current formatted first date [metadata] to display */
//  char last_date[DATE_LENGTH + 1];                        /* Current formatted last date [metadata] to display */
//  char node_date[DATE_LENGTH + 1];                        /* Current formatted node longitude date [calculated] to display */

  time_t first_time_t = (time_t) -1;                      /* */
  time_t last_time_t = (time_t) -1;                       /* */
  time_t node_time_t = (time_t) -1;                       /* */

  const char *format;                                     /* */

  int lat_index=-1;                              /* */
  int lon_index=-1;                              /* */
  int tim_index=-1;                              /* */

  int lat_start_index=-1;                              /* */
  int lon_start_index=-1;                              /* */
  int tim_start_index=-1;                              /* */

  int lat_stop_index=-1;                              /* */
  int lon_stop_index=-1;                              /* */
  int tim_stop_index=-1;                              /* */

  int flag_start_stop=TRUE;
  int flag_nlf=FALSE;

  int fcf_index = -1;
  int nlf_index = -1;

  int sds_index_requested_tab[NMAX_SDS];                   /* Index of SDS for the specified HDF file */
  int sds_index_required_tab[NMAX_SDS];                   /* Index of SDS for the specified HDF file */


  char lat_name[H4_MAX_NC_NAME]       = "Latitude";                       /* */
  char lon_name[H4_MAX_NC_NAME]       = "Longitude";                      /* */

  char lat_start_name[H4_MAX_NC_NAME] = "Latitude_Start";                 /* */
  char lon_start_name[H4_MAX_NC_NAME] = "Longitude_Start";                /* */

  char lat_stop_name[H4_MAX_NC_NAME]  = "Latitude_Stop";                  /* */
  char lon_stop_name[H4_MAX_NC_NAME]  = "Longitude_Stop";                 /* */

  char nlf_name[H4_MAX_NC_NAME] = "Number_Layers_Found";                  /* */

  char tim_name[H4_MAX_NC_NAME]       = "Profile_Time";                   /* */
  char tim_start_name[H4_MAX_NC_NAME] = "Profile_Time_Start";             /* */
  char tim_stop_name[H4_MAX_NC_NAME]  = "Profile_Time_Stop";              /* */

  float node_longitude         = -999.;               /* Value of the longitude(-180.,+180.) corresponding to the ascendant(day)/descendant(night) equator crrosing node */

  float latitude;                                     /* Dummy variable for latitude value for each lidar shots */
  float  *latitude_sds         = NULL;                /* Address corresponding to the Latitude SDS */
  float  *latitude_start       = NULL;                /* Address corresponding to the Latitude SDS */
  float  *latitude_stop        = NULL;                /* Address corresponding to the Latitude SDS */

  float longitude;                                    /* Dummy variable for longitude value for each lidar shots */
  float  *longitude_sds        = NULL;                /* Address corresponding to the Longitude SDS */
  float  *longitude_start      = NULL;                /* Address corresponding to the Longitude SDS */
  float  *longitude_stop       = NULL;                /* Address corresponding to the Longitude SDS */

  double time;                                        /* Dummy variable for time value for each lidar shots */
  double *time_sds             = NULL;                /* Address corresponding to the Profile_Time SDS */
  double *time_start           = NULL;                /* Address corresponding to the Profile_Time SDS */
  double *time_stop            = NULL;                /* Address corresponding to the Profile_Time SDS */

  char   *nlf                  = NULL;                /* Address corresponding to the Number_Layers_Found SDS */

  PTR_SDS *sds_required    = NULL;                /* */
  PTR_SDS *sds_requested   = NULL;                /* */

  int day;                                            /* current day of month */
  int msec;                                           /* current milliseconds of day */

  int jj;

  int32   file_id, vdata_id;
  intn    statusn;
  int32   status32;
  int32   index;
     
  int32   n_records, interlace, vdata_ref;
  void   *buf;

  /* */
  program_name = argv[0];

  if (argc < 2) usage(EXIT_FAILURE);

  /* ARGUMENTS */
  ARGS *args = NULL;
  Malloc(args, sizeof *args);

  /* Decoding arguments */
  if (decodeargs(&argc, &argv, args))
  {
    free(args);
    fatal("Decoding arguments %s:%d",__FILE__,__LINE__);
  }

  /* Get input filename */
  if (argc > 1 ) { strcpy(filein,argv[1]); }
  else           { fatal("HDF input file is mandatory %s:%d",__FILE__,__LINE__); }

  nb_sds=get_hdf_file_nb_sds(filein); /**/
  nb_rec=get_caliop_file_nb_record(filein); /**/

  /* Get the full list of SDS name for the requested input HDF file */
  get_list_SDSs(filein);


  if (argc == 2) /* the only argument is a HDF filename, without parameters to read */
  {
    /* Display global or sds attributes informations */
    if (args->attributes)
    {
      printf("Displaying global or sds attributes shall come soon\n");

      exit(EXIT_SUCCESS);
    }


    /* Display metadata informations */
    if (args->metadata)
    {
      /* Open the HDF file. */
      file_id = Hopen(filein, DFACC_READ, 0);

      /* Initialize the Vset interface. */
      status32 = Vstart(file_id);

      /* Get the reference number for the first Vdata in the file. */
      vdata_ref = -1; /*vdata_ref = VSgetid(file_id, vdata_ref);*/

      /* Get the reference number of the vdata */
      vdata_ref = VSfind(file_id, "metadata");
//      vdata_ref = VSfind(file_id, "INVENTORYMETADATA");
//      if (vdata_ref == 0) { fatal("VSfind vdata_ref:[%d] %s:%d",(int)vdata_ref,__FILE__,__LINE__); }

      /* Attach to the vdata for reading if it is found, otherwise exit the program. */
      vdata_id = VSattach(file_id, vdata_ref, "r");
//      vdata_id = VSattach(file_id, 0, "r");
//printf("vdata_id:[%d]\n",vdata_id);
      if (vdata_id == 0) { fatal("VSattach vdata_id:[%d] %s:%d",(int)vdata_id,__FILE__,__LINE__); }

      if (args->metadata_field) 
      {
        /* Specify the fields that will be read. */
        if (VSsetfields(vdata_id,args->metadata_fieldname) == FAIL) { fatal("VSsetfields:[%s] vdata_id:[%d] %s:%d",args->metadata_fieldname,(int)vdata_id,__FILE__,__LINE__); }

        /* Place the current point to the position specified in RECORD_INDEX. */
        /* record_pos = VSseek (vdata_id, RECORD_INDEX); */
        /* record_pos = VSseek (vdata_id, RECORD_INDEX); */
        /* int32 record_pos = VSseek(vdata_id, 0); */
   
        /* intn VSfindex(int32 vdata_id, char *fieldname, int32 *field_index) */ /* Retrieves the index of a field within a vdata. */
        statusn = VSfindex(vdata_id, args->metadata_fieldname, &index);

        /* int32 VFfieldtype(int32 vdata_id, int32 field_index) */ /* Returns the data type of a vdata field. */
        int32 data_type = VFfieldtype(vdata_id, index);
        /* int32 VFfieldorder(int32 vdata_id, int32 field_index) */ /* Returns the order of a vdata field. */
        int32 data_order = VFfieldorder(vdata_id, index);
        /* int32 VFfieldisize(int32 vdata_id, int32 field_index) */ /* Returns the size, as stored in memory, of a vdata field. */
        //int32 data_msize = VFfieldisize(vdata_id, index);
        /* int32 VFfieldesize(int32 vdata_id, int32 field_index) */ /* Returns the size, as stored on disk, of a vdata field. */
        //int32 data_dsize = VFfieldesize(vdata_id, index);

        /* int32 VSfnattrs(int32 vdata_id, int32 field_index) */ /* Returns the number of attributes attached to a vdata or the number of attributes attached to a vdata field. */
        /* int32 n_attrs = VSfnattrs(vdata_id, index); */

        char cstr_sds_data_type[8];         /* */

        switch(data_type)
        {
          case DFNT_UCHAR  : strcpy(cstr_sds_data_type,"UCHAR");  break;
          case DFNT_CHAR   : strcpy(cstr_sds_data_type,"CHAR");   break;
          case DFNT_UCHAR16: strcpy(cstr_sds_data_type,"UCHAR16"); break;
          case DFNT_CHAR16 : strcpy(cstr_sds_data_type,"CHAR16");  break;
          case DFNT_UINT8  : strcpy(cstr_sds_data_type,"UINT8");   break;
          case DFNT_INT8   : strcpy(cstr_sds_data_type,"INT8");    break;
          case DFNT_UINT16 : strcpy(cstr_sds_data_type,"UINT16");  break;
          case DFNT_INT16  : strcpy(cstr_sds_data_type,"INT16");   break;
          case DFNT_UINT32 : strcpy(cstr_sds_data_type,"UINT32");  break;
          case DFNT_INT32  : strcpy(cstr_sds_data_type,"INT32");   break;
          case DFNT_UINT64 : strcpy(cstr_sds_data_type,"UINT64");  break;
          case DFNT_INT64  : strcpy(cstr_sds_data_type,"INT64");   break;
          case DFNT_UINT128: strcpy(cstr_sds_data_type,"UINT128"); break;
          case DFNT_INT128 : strcpy(cstr_sds_data_type,"INT128");  break;
          case DFNT_FLOAT32: strcpy(cstr_sds_data_type,"FLOAT32"); break;
          case DFNT_FLOAT64: strcpy(cstr_sds_data_type,"FLOAT64"); break;
          default            : fatal("Unknown SDS data type [%d] %s:%d",(int)data_type,__FILE__,__LINE__); break;
        } /* switch(data_type) */
  
        /* intn VSQueryinterlace(int32 vdata_id, int32 *interlace_mode) */ /* Retrieves the interlace mode of the vdata. */
        statusn = VSQueryinterlace(vdata_id,&interlace);

        /* intn VSQuerycount(int32 vdata_id, int32 *n_records) */ /* Retrieves the number of records in a vdata. */
        statusn = VSQuerycount(vdata_id,&n_records);

        size_t size = (size_t) data_order;
        /* Memory allocation for each SDS record */
        switch(data_type)
        {
          case DFNT_UCHAR   :
          case DFNT_UINT8   : Malloc(buf, size * sizeof(unsigned char));          break;
          case DFNT_CHAR    :
          case DFNT_INT8    : Malloc(buf, size * sizeof(char));                   break;
          case DFNT_UINT16  : Malloc(buf, size * sizeof(unsigned short int));     break;
          case DFNT_INT16   : Malloc(buf, size * sizeof(short int));              break;
          case DFNT_UINT32  : Malloc(buf, size * sizeof(unsigned int));           break;
          case DFNT_INT32   : Malloc(buf, size * sizeof(int));                    break;
          case DFNT_UINT64  : Malloc(buf, size * sizeof(unsigned long int));      break;
          case DFNT_INT64   : Malloc(buf, size * sizeof(long int));               break;
          case DFNT_UINT128 : Malloc(buf, size * sizeof(unsigned long long int)); break;
          case DFNT_INT128  : Malloc(buf, size * sizeof(long long int));          break;
          case DFNT_FLOAT32 : Malloc(buf, size * sizeof(float));                  break;
          case DFNT_FLOAT64 : Malloc(buf, size * sizeof(double));                 break;
          default             : fatal("Unknown SDS buf type %s:%d",__FILE__,__LINE__);
        } /* switch(type) */


        /* int32 VSread(int32 vdata_id, uint8 *databuf, int32 n_records, int32 interlace_mode) */ /* Retrieves data from a vdata. */
        status32 = VSread(vdata_id, (uint8*)buf, n_records, interlace);

        format=hdf_format_string((int32)data_type);

          for(int idata=0; idata<data_order; idata++)
          {
            output_line[0] = '\0'; /* Initialize output_line to an empty string */

            if ( (data_type != DFNT_UCHAR8) && (data_type != DFNT_CHAR8) )
            {
              if (args->record_index) { sprintf(token, "%d\t", idata); strcat(output_line, token); }
            } /* Output */

            /* Switch for output each SDS record */
            switch(data_type)
            {
              case DFNT_UCHAR8  :
              case DFNT_UINT8   : { unsigned char *data =                   (unsigned char*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_CHAR8   :
              case DFNT_INT8    : { char *data =                                     (char*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_UINT16  : { unsigned short int *data =         (unsigned short int*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_INT16   : { short int *data =                           (short int*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_UINT32  : { unsigned int *data =                     (unsigned int*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_INT32   : { int *data =                                       (int*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_UINT64  : { unsigned long int *data =           (unsigned long int*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_INT64   : { long int *data =                             (long int*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_UINT128 : { unsigned long long int *data = (unsigned long long int*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_INT128  : { long long int *data =                   (long long int*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_FLOAT32 : { float *data =                                   (float*)buf; sprintf(token, format, data[idata]); break; }
              case DFNT_FLOAT64 : { double *data =                                 (double*)buf; sprintf(token, format, data[idata]); break; }
              default             : free(buf); fatal("Unknown SDS data type %s:%d",__FILE__,__LINE__); break;
            } /* switch */

            strcat(output_line, token);

            if ( (data_type == DFNT_UCHAR8) || (data_type == DFNT_CHAR8) ) { printf("%s", output_line); } /* Output */
            else                                                           { printf("%s\n", output_line); }

          } /* for(int idata=0; idata<data_order; idata++) */

          if ( (data_type == DFNT_UCHAR8) || (data_type == DFNT_CHAR8) ) { printf("\n"); }
        
          free(buf);

      } else {

        /* int32 VFnfields(int32 vdata_id) */ /* Returns the total number of fields in a vdata. */
        int32 n_fields = VFnfields(vdata_id);

        for (int index=0; index<n_fields; index++)
        {
          /* char *VFfieldname(int32 vdata_id, int32 field_index) */ /* Returns the name of a vdata field. */
          char *field_name = VFfieldname(vdata_id, index);

          /* intn VSfindex(int32 vdata_id, char *fieldname, int32 *field_index) */ /* Retrieves the index of a field within a vdata. */
          /*statusn = VSfindex(vdata_id, field_name, &index);*/

          /* int32 VFfieldtype(int32 vdata_id, int32 field_index) */ /* Returns the data type of a vdata field. */
          int32 data_type = VFfieldtype(vdata_id, index);
          /* int32 VFfieldorder(int32 vdata_id, int32 field_index) */ /* Returns the order of a vdata field. */
          int32 data_order = VFfieldorder(vdata_id, index);
          /* int32 VFfieldisize(int32 vdata_id, int32 field_index) */ /* Returns the size, as stored in memory, of a vdata field. */
          //int32 data_msize = VFfieldisize(vdata_id, index);
          /* int32 VFfieldesize(int32 vdata_id, int32 field_index) */ /* Returns the size, as stored on disk, of a vdata field. */
          //int32 data_dsize = VFfieldesize(vdata_id, index);

          /* int32 VSfnattrs(int32 vdata_id, int32 field_index) */ /* Returns the number of attributes attached to a vdata or the number of attributes attached to a vdata field. */
          int32 n_attrs = VSfnattrs(vdata_id, index);

          char cstr_sds_data_type[8];         /* */

          switch(data_type)
          {
            case (DFNT_UCHAR)  : strcpy(cstr_sds_data_type,"UCHAR");   break;
            case (DFNT_CHAR)   : strcpy(cstr_sds_data_type,"CHAR");    break;
            case (DFNT_UINT8)  : strcpy(cstr_sds_data_type,"UINT8");   break;
            case (DFNT_INT8)   : strcpy(cstr_sds_data_type,"INT8");    break;
            case (DFNT_UINT16) : strcpy(cstr_sds_data_type,"UINT16");  break;
            case (DFNT_INT16)  : strcpy(cstr_sds_data_type,"INT16");   break;
            case (DFNT_UINT32) : strcpy(cstr_sds_data_type,"UINT32");  break;
            case (DFNT_INT32)  : strcpy(cstr_sds_data_type,"INT32");   break;
            case (DFNT_UINT64) : strcpy(cstr_sds_data_type,"UINT64");  break;
            case (DFNT_INT64)  : strcpy(cstr_sds_data_type,"INT64");   break;
            case (DFNT_UINT128): strcpy(cstr_sds_data_type,"UINT128"); break;
            case (DFNT_INT128) : strcpy(cstr_sds_data_type,"INT128");  break;
            case (DFNT_FLOAT32): strcpy(cstr_sds_data_type,"FLOAT32"); break;
            case (DFNT_FLOAT64): strcpy(cstr_sds_data_type,"FLOAT64"); break;
            default            : fatal("Unknown SDS data type [%d] %s:%d",(int)data_type,__FILE__,__LINE__); break;
          } /* switch(data_type) */
  
          /* Output record index */

          printf("[%2d]:%55s   data_type:%8s    rank:%2d   n_attrs:%2d  ",index,field_name,cstr_sds_data_type,1,(int)n_attrs);
          printf("dimsizes[00]:%5d  ",(int)data_order);
          printf("\n");
  
        } /* for (int index=0; index<n_fields; index++) */

      } /* if (args->metadata_field) */


      /* Detach from the Vdata, close the interface and the file. */
      status32 = VSdetach(vdata_id);
      status32 = Vend(file_id);
      status32 = Hclose(file_id);

      exit(EXIT_SUCCESS);

    } /* if (args->metadata) */


    /* Display date or time first, node and last acquisition lidar shoot informations */
    if (args->coord_time || args->coord_date)
    {
      struct tm first_tm;
      struct tm last_tm;
      struct tm node_tm;

      int day;
      int msec;
      double tim;

      tim  = get_caliop_file_first_time(filein);

      day  = (int) (tim/86400.);
      msec = (int) (86400000. * (tim/86400. - day));
      cds2jmahms(day,msec,&first_tm.tm_mday,&first_tm.tm_mon,&first_tm.tm_year,&first_tm.tm_hour,&first_tm.tm_min,&first_tm.tm_sec);

      tim  = get_caliop_file_node_time(filein);
      day  = (int) (tim/86400.);
      msec = (int) (86400000. * (tim/86400. - day));
      cds2jmahms(day,msec,&node_tm.tm_mday,&node_tm.tm_mon,&node_tm.tm_year,&node_tm.tm_hour,&node_tm.tm_min,&node_tm.tm_sec);

      tim  = get_caliop_file_last_time(filein);
      day  = (int) (tim/86400.);
      msec = (int) (86400000. * (tim/86400. - day));
      cds2jmahms(day,msec,&last_tm.tm_mday,&last_tm.tm_mon,&last_tm.tm_year,&last_tm.tm_hour,&last_tm.tm_min,&last_tm.tm_sec);

      first_tm.tm_year -= 1900;
      last_tm.tm_year -= 1900;
      node_tm.tm_year -= 1900;

      first_tm.tm_mon--;
      last_tm.tm_mon--;
      node_tm.tm_mon--;

      first_tm.tm_isdst = 0;
      last_tm.tm_isdst = 0;
      node_tm.tm_isdst = 0;

      first_time_t = my_timegm(&first_tm);
      last_time_t = my_timegm(&last_tm);
      node_time_t = my_timegm(&node_tm);

      if (! args->time_format_is_set)
      {
        strncpy(args->time_format, "%F\t%H:%M:%S", sizeof(args->time_format) - 1);
      }

      strftime(cstr_first_time, sizeof(cstr_first_time), args->time_format, &first_tm);
      strftime(cstr_last_time, sizeof(cstr_last_time), args->time_format, &last_tm);
      strftime(cstr_node_time, sizeof(cstr_node_time), args->time_format, &node_tm);

      fprintf(stdout, "%s\t%s\t%s\n", cstr_first_time, cstr_node_time, cstr_last_time);

      exit(EXIT_SUCCESS);

    } /* if (args->coord_time) */


    /* Display node longitude informations */
    if (args->node_longitude)
    {
      lat_start_index = get_index_SDS(lat_start_name,nb_sds);
      if ( lat_start_index == FAIL ) { flag_start_stop = FALSE; }

      node_longitude = get_caliop_file_node_longitude(filein,flag_start_stop);
      if ( node_longitude > -180. ) { fprintf(stdout, "%.5f\n", node_longitude); }
      else                          { fatal("Get node longitude value %s:%d",__FILE__,__LINE__); }
      exit(EXIT_SUCCESS);

    } /* if (args->node_longitude) */

    /* Test whether user have request for all SDS. */
    if ((args->coord_polder) || (args->coord_latlon))
    {
    }
    else if (args->allsds)
    {
    } else {
      print_list_SDSs(filein);
      exit(EXIT_SUCCESS);
    }
  } /* if (argc == 2) */

  /* Check that the HDF input filename has the .hdf or .HDF extension (mandatory) */
  if ( (strcasestr(filein,".hdf") == NULL) )
  {
    fatal("1The input file [%s] should have a .hdf or .HDF extension %s:%d",filein,__FILE__,__LINE__);
  } /**/


  int ind=0;

  lat_start_index = get_index_SDS(lat_start_name,nb_sds);
  if ( lat_start_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],lat_start_name);
    sds_index_required_tab[ind] = lat_start_index;
    lat_start_index = ind; ind++; assert(ind<NMAX_SDS);
  }

  lat_stop_index = get_index_SDS(lat_stop_name,nb_sds);
  if ( lat_stop_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],lat_stop_name);
    sds_index_required_tab[ind] = lat_stop_index;
    lat_stop_index = ind; ind++; assert(ind<NMAX_SDS);
  }

  lon_start_index = get_index_SDS(lon_start_name,nb_sds);
  if ( lon_start_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],lon_start_name);
    sds_index_required_tab[ind] = lon_start_index;
    lon_start_index = ind; ind++; assert(ind<NMAX_SDS);
  }

  lon_stop_index = get_index_SDS(lon_stop_name,nb_sds);
  if ( lon_stop_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],lon_stop_name);
    sds_index_required_tab[ind] = lon_stop_index;
    lon_stop_index = ind; ind++; assert(ind<NMAX_SDS);
  }

  tim_start_index = get_index_SDS(tim_start_name,nb_sds);
  if ( tim_start_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],tim_start_name);
    sds_index_required_tab[ind] = tim_start_index;
    tim_start_index = ind; ind++; assert(ind<NMAX_SDS);
  }

  tim_stop_index = get_index_SDS(tim_stop_name,nb_sds);
  if ( tim_stop_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],tim_stop_name);
    sds_index_required_tab[ind] = tim_stop_index;
    tim_stop_index = ind; ind++; assert(ind<NMAX_SDS);

    flag_start_stop=TRUE;
  }

  lat_index = get_index_SDS(lat_name,nb_sds);
  if ( lat_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],lat_name);
    sds_index_required_tab[ind] = lat_index;
    lat_index = ind; ind++; assert(ind<NMAX_SDS);
  }

  lon_index = get_index_SDS(lon_name,nb_sds);
  if ( lon_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],lon_name);
    sds_index_required_tab[ind] = lon_index;
    lon_index = ind; ind++; assert(ind<NMAX_SDS);
  }

  tim_index = get_index_SDS(tim_name,nb_sds);
  if ( tim_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],tim_name);
    sds_index_required_tab[ind] = tim_index;
    tim_index = ind; ind++; assert(ind<NMAX_SDS);

    flag_start_stop=FALSE;
  }

  nlf_index = get_index_SDS(nlf_name,nb_sds);
 
  if ( nlf_index != FAIL )
  {
    strcpy(sds_name_required_tab[ind],nlf_name);
    sds_index_required_tab[ind] = nlf_index;
    nlf_index = ind; ind++; assert(ind<NMAX_SDS);
    flag_nlf = TRUE;
  }
 


  nb_sds_required=ind;

  /* Dynamic memory allocation for output data */
  sds_required = newPTR_SDSs(filein, sds_index_required_tab, nb_sds_required);
  assert(sds_required);

  nb_sds_requested=argc-2;
  assert(nb_sds_requested<NMAX_SDS);

  /* Test whether user have request for all SDS. */
  if ( (args->allsds) ) {
    nb_sds_requested = nb_sds;
    for (k=0;k<nb_sds_requested;k++) { strcpy(sds_name_requested_tab[k], SDS_names[k]); }
  } else {
    for (k=0;k<nb_sds_requested;k++) { strcpy(sds_name_requested_tab[k], argv[k+2]); }
  }


  /* Test whether user have request for specific's Feature_Classification_Flags SDS. */
  char *buf_fcf = "Feature_Classification_Flags";
  for (int k=0;k<nb_sds_requested;k++)
  {
    if ((strncmp(sds_name_requested_tab[k], buf_fcf, strlen(buf_fcf)) == 0) )
    {
      fcf_index = k;
    }
  }


  /* Check that the requested products are available */
  if ((argc > 2) || (args->allsds))
  {
    int index_requested=0;                          /* Index of the current sds required */

    for (k=0;k<nb_sds_requested;k++)
    {
      assert(k<NMAX_SDS);

      //strcpy(sds_name, argv[k+2]);
      strcpy(sds_name, sds_name_requested_tab[k]);

      index_requested=get_index_SDS(sds_name,nb_sds);
      if ( index_requested == FAIL)
      { 
        error("SDS [%s] is unavailable in file [%s]\n",sds_name,filein);
        error("You should invoke %s only with mandatory filename argument to list all available SDS's as:",program_name);
        error("%s %s\n\n",program_name,filein);
        exit(EXIT_FAILURE);
      }
      else
      {
        sds_index_requested_tab[k] = index_requested;
      }

    } /* for (k=0;k<argc-2;k++) */

  } /* if (argc > 2) */


  /* Dynamic memory allocation for output data */
  sds_requested = newPTR_SDSs(filein, sds_index_requested_tab, nb_sds_requested);
  assert(sds_requested);


  /* Memory allocation for required sds data */

  if ( flag_start_stop == TRUE )
  {
    if ( lat_start_index != FAIL ) { latitude_start = (float*)(sds_required[lat_start_index].data); assert(latitude_start); }
    if ( lat_stop_index != FAIL )  { latitude_stop  = (float*)(sds_required[lat_stop_index].data);  assert(latitude_stop); }

    if ( lon_start_index != FAIL ) { longitude_start = (float*)(sds_required[lon_start_index].data); assert(longitude_start); }
    if ( lon_stop_index != FAIL )  { longitude_stop  = (float*)(sds_required[lon_stop_index].data);  assert(longitude_stop); }

    if ( tim_start_index != FAIL ) { time_start = (double*)(sds_required[tim_start_index].data); assert(time_start); }
    if ( tim_stop_index != FAIL )  { time_stop  = (double*)(sds_required[tim_stop_index].data);  assert(time_stop); }
  }
  else
  {
    if ( lat_index != FAIL ) { latitude_sds  = (float*)(sds_required[lat_index].data);  assert(latitude_sds); }

    if ( lon_index != FAIL ) { longitude_sds = (float*)(sds_required[lon_index].data);  assert(longitude_sds); }

    if ( tim_index != FAIL ) { time_sds      = (double*)(sds_required[tim_index].data); assert(time_sds); }
  }

  nlf = (char*)(sds_required[nlf_index].data);

/*
  separator = "\t";
  sprintf(output_format,"%s%s",output_format_temp, separator);
*/


  /* Loop over pixels */
  for (j=0;j<nb_rec;j++)
  {
    output_line[0] = '\0'; /* Initialize output_line to an empty string */

    /* Output index if required by user */
    int irec = j + args->language_offset;


    /* Output record index */
    if ((args->record_index)) { sprintf(token, "%d\t", irec); strcat(output_line, token); }


    /* Output line/column values */
    if ( (args->coord_polder) || (args->coord_latlon) )
    {
      jj = j;

      if ( flag_start_stop == TRUE )
      {
        if (sds_required[lat_start_index].dimsizes[1] > 1) jj = 1 + sds_required[lat_start_index].dimsizes[1] * j;

        latitude  = ( latitude_start[jj]  + latitude_stop[jj]  ) / 2.;

        /* Here, we test if longitude_start and longitude_stop have opposite signs.
           If so (Hope to be near 180 degres !), we take the mean of the absolute values. */
        if ( ((longitude_start[jj] > 0.) && (longitude_stop[jj] < 0.)) ||
             ((longitude_start[jj] < 0.) && (longitude_stop[jj] > 0.)) ) 
        { 
          longitude = ( fabsf(longitude_start[jj]) + fabsf(longitude_stop[jj]) ) / 2.;
        } 
        else
        {
          longitude = ( longitude_start[jj] + longitude_stop[jj] ) / 2.;
        }
      } 
      else
      {
        if (sds_required[lat_index].dimsizes[1] > 1) jj = 1 + sds_required[lat_index].dimsizes[1] * j;
   
        latitude  = latitude_sds[jj];
        longitude = longitude_sds[jj];
    
      }
    } /* if ( args->coord_polder || args->coord_latlon ) */


    if ( (args->coord_polder) )
    {
      float flig,fcol;
      if ( latlon2ligcol_polder(latitude,longitude,args->resolution,&flig,&fcol) == FAIL )
      {
        error("Error in computing latlon->ligcol");
        if (sds_required)  freePTR_SDSs(sds_required,nb_sds_required);
        if (sds_requested) freePTR_SDSs(sds_requested,nb_sds_requested);
        exit(2);
      } 
      else
      {
        lig=NINT(flig);
        col=NINT(fcol);
      }
      if ( (lig > args->min_line)   && (lig < args->max_line) &&
           (col > args->min_column) && (col < args->max_column) )
      {
        if (args->reverse_coord) { sprintf(token, "%d\t%d\t", col, lig); strcat(output_line, token); }
        else                         { sprintf(token, "%d\t%d\t", lig, col); strcat(output_line, token); }
      }
      else
      {
        continue;
      }
    } /* if (args->coord_polder) */


    /* Output Latitude/longitude values */
    if ( args->coord_latlon )
    {
      if ( (latitude  > args->min_latitude)  && (latitude  < args->max_latitude) &&
           (longitude > args->min_longitude) && (longitude < args->max_longitude) )
      {
        if (args->reverse_coord) { sprintf(token, "%9.6f\t%9.6f\t", longitude, latitude); }
        else                         { sprintf(token, "%9.6f\t%9.6f\t", latitude, longitude); }
        strcat(output_line, token);
      }
      else
      {
        continue;
      } /* Test over latitude and longitude ranges */
    } /* if ( args->coord_latlon ) */


    /* Output date values */
    if ( args->coord_date)
    {
      jj = j;

      time_t current_time_t = (time_t) -1;          /* */
      struct tm current_tm;

      if ( flag_start_stop == TRUE )
      {
        if (sds_required[tim_start_index].dimsizes[1] > 1) jj = 1 + sds_required[tim_start_index].dimsizes[1] * j;

        time = ( time_start[jj]  + time_stop[jj]  ) / 2.;

      } 
      else
      {
        if (sds_required[tim_index].dimsizes[1] > 1) jj = 1 + sds_required[tim_index].dimsizes[1] * j;
   
        time  = time_sds[jj];
    
      }

      day  = (int) (time/86400.);
      msec = (int) (86400000. * (time/86400. - day));

      cds2jmahms(day,msec,&current_tm.tm_mday,&current_tm.tm_mon,&current_tm.tm_year,&current_tm.tm_hour,&current_tm.tm_min,&current_tm.tm_sec);

      current_tm.tm_year -= 1900;
      current_tm.tm_mon--;
      current_tm.tm_isdst = 0;

      current_time_t = my_timegm(&current_tm);

      strftime(cstr_current_time, sizeof(cstr_current_time), args->date_format, gmtime(&current_time_t));
      sprintf(token, "%s\t", cstr_current_time);
      strcat(output_line, token);
    } /* if ( args->coord_time ) */


    /* Output time values */
    if ( args->coord_time)
    {
      jj = j;

      time_t current_time_t = (time_t) -1;          /* */
      struct tm current_tm;

      if ( flag_start_stop == TRUE )
      {
        if (sds_required[tim_start_index].dimsizes[1] > 1) jj = 1 + sds_required[tim_start_index].dimsizes[1] * j;

        time = ( time_start[jj]  + time_stop[jj]  ) / 2.;

      } 
      else
      {
        if (sds_required[tim_index].dimsizes[1] > 1) jj = 1 + sds_required[tim_index].dimsizes[1] * j;
   
        time  = time_sds[jj];
    
      }

      day  = (int) (time/86400.);
      msec = (int) (86400000. * (time/86400. - day));

      cds2jmahms(day,msec,&current_tm.tm_mday,&current_tm.tm_mon,&current_tm.tm_year,&current_tm.tm_hour,&current_tm.tm_min,&current_tm.tm_sec);

      current_tm.tm_year -= 1900;
      current_tm.tm_mon--;
      current_tm.tm_isdst = 0;

      current_time_t = my_timegm(&current_tm);

      strftime(cstr_current_time, sizeof(cstr_current_time), args->time_format, gmtime(&current_time_t));
      sprintf(token, "%s\t", cstr_current_time);
      strcat(output_line, token);
    } /* if ( args->coord_time ) */


    if ((argc > 2) || (args->allsds))
    {
      /* Loop over available requested SDS */
      for (int index_sds=0;index_sds<nb_sds_requested;index_sds++)
      {
        int array_index[2];                            /* Array index bounds for the current SDS to display */

        array_index[0]  = 0;
        array_index[1]  = sds_requested[index_sds].dimsizes[1] - 1;

        array_index[0] += args->language_offset;
        array_index[1] += args->language_offset;

        int bounds_flag = get_arrays_bound_index_SDS(sds_name_requested_tab[index_sds],&array_index[0]);

        /* Test on lower/upper array bounds index */
        if ( array_index[0] > array_index[1] ) { fatal("Lower bound array index is greater than upper bound array index [%s] %s:%d",sds_requested[index_sds].name,__FILE__,__LINE__); }
        if ( (array_index[1]+1-args->language_offset) > sds_requested[index_sds].dimsizes[1] )
        {
          fatal("Required upper bound is inconsistent with the dimsize for SDS [%s] %s:%d",sds_requested[index_sds].name,__FILE__,__LINE__);
        }
        if ( (array_index[0]-args->language_offset) < 0 )
        {
          fatal("Required lower bound is inconsistent with the dimsize for SDS [%s] %s:%d",sds_requested[index_sds].name,__FILE__,__LINE__);
        }

        /* Display only Number_Layers_Found's SDS values if user request it ([+]) */
	if (flag_nlf && (bounds_flag == BOUNDS_NLF) )
	{
	  int idata_nlf = sds_required[nlf_index].dimsizes[1] * j;

          array_index[0]  = 0;
          array_index[1]  = nlf[idata_nlf] - 1;

          array_index[0] += args->language_offset;
          array_index[1] += args->language_offset;
	}

        /* Display Feature_Classification_Flags specific's SDS */
        if ( (args->decode_fcf > 0) && (index_sds == fcf_index) )
        {
          display_fcf(j, sds_requested, index_sds, args->language_offset, args->decode_fcf, &array_index[0], output_line);

        } else {

          format=hdf_format_string((int32)sds_requested[index_sds].type);

          /* Loop over array elements */
          for (int l=array_index[0];l<=array_index[1];l++)
          {
            int idata=(l-args->language_offset) + sds_requested[index_sds].dimsizes[1] * j;
  
            /* Switch for output each SDS record */
            switch(sds_requested[index_sds].type)
            {
              case (DFNT_UINT8)   : { unsigned char *data =                   (unsigned char*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_INT8)    : { char *data =                                     (char*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_UINT16)  : { unsigned short int *data =         (unsigned short int*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_INT16)   : { short int *data =                           (short int*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_UINT32)  : { unsigned int *data =                     (unsigned int*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_INT32)   : { int *data =                                       (int*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_UINT64)  : { unsigned long int *data =           (unsigned long int*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_INT64)   : { long int *data =                             (long int*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_UINT128) : { unsigned long long int *data = (unsigned long long int*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_INT128)  : { long long int *data =                   (long long int*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_FLOAT32) : { float *data =                                   (float*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              case (DFNT_FLOAT64) : { double *data =                                 (double*)(sds_requested[index_sds].data); sprintf(token, format, data[idata]); break; }
              default             : fatal("Unknown SDS data type %s:%d",__FILE__,__LINE__); break;
            } /* switch */

            /**/
            strcat(output_line, token);

          } /* for (int l=lmin;l<lmax;l++) */

        }

      } /* for (int index_sds=0;index_sds<argc-2;index_sds++) */

    } else { /* Degraded case where no SDS name are required, but latitude and longitude (-g option) is passed through arguments */
    } /* if (argc > 2) */

    /* Output line */
    if (output_line[0] != '\0') { fprintf(stdout,"%s\n", output_line); } /* Output */

  } /* End loop over pixels */


  /* Free memory */
  if (sds_required)  { freePTR_SDSs(sds_required,nb_sds_required); }
  if (sds_requested) { freePTR_SDSs(sds_requested,nb_sds_requested); }

  if (args) { free(args); }

  exit(EXIT_SUCCESS);
} /* int main(int argc, char **argv) */
